﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.Threading;
using System.Runtime.InteropServices;
using System.IO;

using PlatformSDKSample;

namespace LiveSnapshot
{
    using TStreamInfo = PlatformSDK.TStreamInfo;
    using TVideoDecodeInfo = PlatformSDK.TVideoDecodeInfo;
    using TStreamingDisplayInfo = PlatformSDK.TStreamingDisplayInfo;

    public partial class ViewCell : Form
    {
        IntPtr hLiveChannel = IntPtr.Zero;
        IntPtr hPlaybackChannel = IntPtr.Zero;
        readonly string mode = "";

        public ViewCell(string name)
        {
            mode = name;
            InitializeComponent(name);
        }

        private void VideoDecodeCallBack(IntPtr pVideoDecodeInfo)
        {
            EventLog.Write("*** VideoDecodeCallBack ***");
            TVideoDecodeInfo decodeInfo = (TVideoDecodeInfo)Marshal.PtrToStructure(pVideoDecodeInfo, typeof(TVideoDecodeInfo));

            if (decodeInfo.pBuffer == IntPtr.Zero)
            {
                EventLog.Write("Decode buffer is empty");
                return;
            }

            SaveImage(ref decodeInfo);

            Thread t1 = new Thread(doDeleteChannel);
            t1.Start();

            Action<bool> updateButton1Delegate = (enabled) => { this.button1.Enabled = enabled; };
            this.Invoke(updateButton1Delegate, true);
        }

        unsafe void SaveImage(ref TVideoDecodeInfo decodeInfo)
        {
            if (decodeInfo.pBuffer == IntPtr.Zero) return;

            System.DateTime time = new System.DateTime(1970, 1, 1).AddSeconds(decodeInfo.frameSecond);
            string savePath = "D:\\Export1\\" + time.ToString("yyyyMMdd_HHmmss")
                + "_" + decodeInfo.frameMilSec + ".jpg";

            EventLog.Write("SaveImage size:{0}", decodeInfo.bufSize);
            try
            {
                byte[] image = new byte[decodeInfo.bufSize];
                Marshal.Copy(decodeInfo.pBuffer, image, 0, decodeInfo.bufSize);
                FileStream file = new FileStream(savePath, FileMode.OpenOrCreate, FileAccess.Write);
                file.Write(image, 0, decodeInfo.bufSize);
                file.Close();

                EventLog.Write("SaveImage finished !!!");
            }
            catch (IOException e)
            {
                EventLog.Write("Fail to save image \"" + savePath + "\":\n" + e.ToString());
            }
        }

        private void SaveSnapshot(object sender, EventArgs e)
        {
            button1.Enabled = false;

            TStreamInfo streamInfo = new TStreamInfo();
            streamInfo.streamSource = PlatformSDK.EStreamSource.VAST;
            streamInfo.streamMode = PlatformSDK.EStreamMode.retrieveOneSnapshot;
            streamInfo.timeType = PlatformSDK.ETimeType.recordingTime;
            streamInfo.serverIP = "10.16.104.43";
            streamInfo.serverUserName = "admin";
            streamInfo.serverPassword = "aaaa1111";
            streamInfo.serverPort = 3454;
            streamInfo.cameraID = "C_1";
            streamInfo.pfVideoDecodeCB = new PlatformSDK.pfVideoDecodeCallBack(VideoDecodeCallBack);
            streamInfo.hDisplayWindow = IntPtr.Zero;
            streamInfo.streamIndex = 1;
            streamInfo.capVersion = PlatformSDK.ECapabilityVersion.VAST;
            streamInfo.streamInitStatus = PlatformSDK.EStreamInitStatus.play;

            if (mode == "Live")
            {
                int ret = PlatformSDK.CreateLiveChannel(ref hLiveChannel, ref streamInfo);
                EventLog.Write("PlatformSDK.CreateLiveChannel Ret = {0:X}", ret);
            }
            else
            {
                streamInfo.startTime = 1586225410;
                streamInfo.endTime = streamInfo.startTime + 86400;
                streamInfo.timeKind = PlatformSDK.ETimeKind.UTC;

                EventLog.Write("Before CreateNetFileChannel");
                int ret = PlatformSDK.CreateNetFileChannel(ref hPlaybackChannel, ref streamInfo);
                EventLog.Write("PlatformSDK.CreateNetFileChannel Ret = {0:X}", ret);
            }
        }

        private void DeleteChannel(object sender, EventArgs e)
        {
            doDeleteChannel();
        }

        void doDeleteChannel()
        {
            if (mode == "Live")
            {
                EventLog.Write("Before DeleteLiveChannel");
                int ret = PlatformSDK.DeleteLiveChannel(ref hLiveChannel);
                EventLog.Write("DeleteLiveChannel: {0:X}", ret);
            }
            else
            {
                EventLog.Write("Before DeleteNetFileChannel");
                int ret = PlatformSDK.DeleteNetFileChannel(ref hPlaybackChannel);
                EventLog.Write("DeleteNetFileChannel: {0:X}", ret); 
            }
        }
    }
}
