﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.Threading;
using System.IO;
using System.Runtime.InteropServices;

using PlatformSDKSample;

namespace FisheyeDewarp
{
    public partial class ViewCell : Form
    {
        IntPtr hChannel = IntPtr.Zero;
        bool mouseLeftDown = false;
        bool playLive = false;
        bool playback = false;
        private static readonly DateTime dateTime1970 = new DateTime(1970, 1, 1);

        Dictionary<int, PlatformSDK.EFisheyeDisplayMode> modeMapping = new Dictionary<int, PlatformSDK.EFisheyeDisplayMode>()
        {
            { 0, PlatformSDK.EFisheyeDisplayMode.mode_1O },
            { 1, PlatformSDK.EFisheyeDisplayMode.mode_1O_Unstretched },
            { 2, PlatformSDK.EFisheyeDisplayMode.mode_1R },
            { 3, PlatformSDK.EFisheyeDisplayMode.mode_1P },
            { 4, PlatformSDK.EFisheyeDisplayMode.mode_2P },
            { 5, PlatformSDK.EFisheyeDisplayMode.mode_1P2R },
            { 6, PlatformSDK.EFisheyeDisplayMode.mode_1O3R },
            { 7, PlatformSDK.EFisheyeDisplayMode.mode_4R },
            { 8, PlatformSDK.EFisheyeDisplayMode.mode_1P3R },
            { 9, PlatformSDK.EFisheyeDisplayMode.mode_1O8R },
            { 10, PlatformSDK.EFisheyeDisplayMode.mode_1P3R_TopPanorama },
            { 11, PlatformSDK.EFisheyeDisplayMode.mode_1P2R_TopPanorama }
        };

        public ViewCell()
        {
            InitializeComponent();
        }

        uint UtcDateTime2time_t(DateTime datetime)
        {
            TimeSpan diff = datetime.ToUniversalTime() - dateTime1970;
            return (uint)(diff.TotalSeconds);
        }

        public void DeleteChannel()
        {
            if (playLive)
            {
                var ret1 = PlatformSDK.DeleteLiveChannel(ref hChannel);
                EventLog.Write("PlatformSDK.DeleteLiveChannel Ret = {0:X}", ret1);
                playLive = false;
            }

            if (playback)
            {
                var ret2 = PlatformSDK.DeleteNetFileChannel(ref hChannel);
                EventLog.Write("PlatformSDK.DeleteNetFileChannel Ret = {0:X}", ret2);
                playback = false;
            }
        }

        void SetDisplayInfo()
        {
            PlatformSDK.TStreamingDisplayInfo info = new PlatformSDK.TStreamingDisplayInfo();

            info.showFrameTime = true;
            info.showMillisecond = true;
            info.showTextShadow = true;
            info.opaqueBackground = true;
            info.timeSource = PlatformSDK.ETimeSource.recordingTime;
            info.timeKind = PlatformSDK.ETimeKind.local;
            info.ARGBTextColor = (255 << 16) | (255 << 8) | 255;
            info.ARGBBackgroundColor = (0 << 16) | (0 << 8) | 0;
            info.fontSize = 22;

            int ret = PlatformSDK.SetDisplayInfo(hChannel, ref info);
            EventLog.Write("SetDisplayInfo: {0:X}", ret);
        }

        public void CreateLiveChannel(ref TCreateChannelInfo info)
        {
            DeleteChannel();
            PlatformSDK.TStreamInfo streamInfo = new PlatformSDK.TStreamInfo();

            streamInfo.streamSource = PlatformSDK.EStreamSource.VAST;
            streamInfo.streamMode = PlatformSDK.EStreamMode.normalStream;
            streamInfo.serverIP = info.serverIP;
            streamInfo.serverUserName = info.serverUserName;
            streamInfo.serverPassword = info.serverPassword;
            streamInfo.serverPort = info.serverPort;
            streamInfo.cameraID = info.cameraID;
            streamInfo.pfStreamStatusCB = null;
            streamInfo.pfStreamDataCB = null;
            streamInfo.pfPCMDataCB = null;
            streamInfo.pfYV12DataCB = null;
            streamInfo.pfVideoDecodeCB = null;
            streamInfo.hDisplayWindow = panel1.Handle;
            streamInfo.streamIndex = info.streamIndex;
            streamInfo.streamInitStatus = PlatformSDK.EStreamInitStatus.play;
            streamInfo.capVersion = info.capVersion;
            streamInfo.useSSL = info.useSSL;
            streamInfo.videoDisplayMode = PlatformSDK.EVideoDisplayMode.fillViewCell;

            var ret = PlatformSDK.CreateLiveChannel(ref hChannel, ref streamInfo);
            EventLog.Write("PlatformSDK.CreateLiveChannel Ret = {0:X}", ret);
            playLive = true;

            SetDisplayInfo();
        }

        public void StartSyncPlayback(ref TCreateChannelInfo info)
        {
            IntPtr hAvSyncPlaybackGroup = IntPtr.Zero;
            int ret11 = PlatformSDK.CreateSyncPlaybackGroup(ref hAvSyncPlaybackGroup);
            EventLog.Write("PlatformSDK.CreateSyncPlaybackGroup Ret = {0:X}", ret11);

            var ret12 = PlatformSDK.InsertChannelToGroup(hAvSyncPlaybackGroup, hChannel);
            EventLog.Write("PlatformSDK.InsertChannelToGroup Ret = {0:X}", ret12);

            var option = new PlatformSDK.TSDKSyncPlaybackOption();
            option.startTime = info.startTime;
            option.endTime = info.endTime;
            option.statusCB = null;
            option.statusContext = IntPtr.Zero;
            option.timeKind = PlatformSDK.ETimeKind.UTC;

            var ret3 = PlatformSDK.StartSyncPlaybackGroup(hAvSyncPlaybackGroup, ref option);
            EventLog.Write("PlatformSDK.StartSyncPlaybackGroup Ret = {0:X}", ret3);
        }

        public void CreateNetFileChannel(ref TCreateChannelInfo info)
        {
            DeleteChannel();
            PlatformSDK.TStreamInfo streamInfo = new PlatformSDK.TStreamInfo();

            streamInfo.streamSource = PlatformSDK.EStreamSource.VAST;
            streamInfo.streamMode = PlatformSDK.EStreamMode.normalStream;
            streamInfo.timeType = PlatformSDK.ETimeType.recordingTime;
            streamInfo.serverIP = info.serverIP;
            streamInfo.serverUserName = info.serverUserName;
            streamInfo.serverPassword = info.serverPassword;
            streamInfo.serverPort = info.serverPort;
            streamInfo.cameraID = info.cameraID;
            streamInfo.pfStreamStatusCB = null;
            streamInfo.pfStreamDataCB = null;
            streamInfo.pfPCMDataCB = null;
            streamInfo.pfYV12DataCB = null;
            streamInfo.pfVideoDecodeCB = null;
            streamInfo.hDisplayWindow = panel1.Handle;
            streamInfo.startTime = info.startTime;
            streamInfo.endTime = info.endTime;
            streamInfo.streamIndex = info.streamIndex;
            streamInfo.streamInitStatus = PlatformSDK.EStreamInitStatus.play;
            streamInfo.timeKind = PlatformSDK.ETimeKind.UTC;
            streamInfo.capVersion = info.capVersion;
            streamInfo.useSSL = info.useSSL;
            streamInfo.videoDisplayMode = PlatformSDK.EVideoDisplayMode.fillViewCell;
            streamInfo.playbackMode = PlatformSDK.EPlaybackMode.synchronous;

            int ret = PlatformSDK.CreateNetFileChannel(ref hChannel, ref streamInfo);
            EventLog.Write("PlatformSDK.CreateNetFileChannel Ret = {0:X}", ret);
            playback = true;

            SetDisplayInfo();
        }

        public bool CheckValidFisheyeCamera()
        {
            const int MAX_RETRY_COUNT = 20;
            int retryCount = 0;

            var mountType = PlatformSDK.EFisheyeMountType.unknown;
            while (mountType == PlatformSDK.EFisheyeMountType.unknown)
            {
                if (retryCount == MAX_RETRY_COUNT) break;
                retryCount++;

                var ret = PlatformSDK.GetFisheyeMountType(hChannel, ref mountType);
                EventLog.Write("PlatformSDK.GetFisheyeMountType Ret = {0:X}, type= {1}", ret, mountType);

                Thread.Sleep(500);
            }

            var isFisheyeCamera = mountType != PlatformSDK.EFisheyeMountType.unknown && mountType != PlatformSDK.EFisheyeMountType.nonFisheyeCamera;
            comboBox1.Enabled = isFisheyeCamera;

            return isFisheyeCamera;
        }

        private void comboBox1_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (hChannel != IntPtr.Zero)
            {
                var ret = PlatformSDK.SetFisheyeDisplayMode(hChannel, modeMapping[comboBox1.SelectedIndex]);
                EventLog.Write("PlatformSDK.SetFisheyeDisplayMode Ret = {0:X} {1}", ret, comboBox1.SelectedIndex);
            }
        }

        private void panel1_MouseDown(object sender, System.Windows.Forms.MouseEventArgs e)
        {
            mouseLeftDown = true;
            HandleMouseEvent(e, PlatformSDK.EMouseEventType.leftDown);
        }

        private void panel1_MouseMove(object sender, System.Windows.Forms.MouseEventArgs e)
        {
            if (mouseLeftDown) HandleMouseEvent(e, PlatformSDK.EMouseEventType.dragging);
        }

        private void panel1_MouseUp(object sender, System.Windows.Forms.MouseEventArgs e)
        {
            mouseLeftDown = false;
            HandleMouseEvent(e, PlatformSDK.EMouseEventType.leftUp);
        }

        private void panel1_MouseWheel(object sender, System.Windows.Forms.MouseEventArgs e)
        {
            HandleMouseEvent(e, PlatformSDK.EMouseEventType.wheelRotation);
        }

        private void layoutChanged(object sender, System.Windows.Forms.LayoutEventArgs e)
        {
            EventLog.Write("MyForm_Layout = {0} {1}", this.Width, this.Height);
            panel1.Width = this.Width - 40;
            panel1.Height = this.Height - 250;
        }

        private void HandleMouseEvent(System.Windows.Forms.MouseEventArgs e, PlatformSDK.EMouseEventType eType)
        {
            PlatformSDK.TMouseEvent mouseEvent = new PlatformSDK.TMouseEvent();
            mouseEvent.eventType = eType;
            mouseEvent.x = e.X;
            mouseEvent.y = e.Y;
            mouseEvent.wheelRotation = e.Delta;

            var ret = PlatformSDK.SetFisheyeMouseEvent(hChannel, ref mouseEvent);
            EventLog.Write("PlatformSDK.SetFisheyeMouseEvent Ret = {0:X}", ret);
        }

        private void fisheyeSnapshot_Click(object sender, EventArgs e)
        {
            PlatformSDK.TSnapshotRawInfo snapshotRawInfo = new PlatformSDK.TSnapshotRawInfo();
            var ret = PlatformSDK.GetFisheyeChannelSnapshot(hChannel, (IntPtr)panel1.Handle, ref snapshotRawInfo);

            EventLog.Write("PlatformSDK.GetFisheyeChannelSnapshot Ret = {0:X}", ret);

            if (snapshotRawInfo.pBuffer == IntPtr.Zero)
            {
                EventLog.Write("FisheyeChannelSnapshot is empty", ret);
                return;
            }

            string savePath = "D:\\FromRawData.jpg";
            label1.Text = "Save snapshot: " + savePath;
            try
            {
                byte[] image = new byte[snapshotRawInfo.bufSize];
                Marshal.Copy(snapshotRawInfo.pBuffer, image, 0, snapshotRawInfo.bufSize);
                FileStream file = new FileStream(savePath, FileMode.OpenOrCreate, FileAccess.Write);
                file.Write(image, 0, snapshotRawInfo.bufSize);
                file.Close();
            }
            catch (IOException err)
            {
                EventLog.Write("Fail to save image \"" + savePath + "\":\n" + err.ToString());
            }

            Marshal.FreeHGlobal(snapshotRawInfo.pBuffer);
        }

        private void resizeButton_Click(object sender, EventArgs e)
        {
            EventLog.Write("PlatformSDK.resizeButton_Click Ret = {0} {1}", this.Height, this.Width);

            panel1.Width = this.ClientSize.Width - 20;
            panel1.Height = this.ClientSize.Height - 200;
        }

        private void button1_Click(object sender, EventArgs e)
        {
            TCreateChannelInfo info = new TCreateChannelInfo();
            info.serverIP = textBox1.Text;
            info.serverUserName = textBox3.Text;
            info.serverPassword = textBox4.Text;
            info.serverPort = Convert.ToInt32(textBox2.Text);
            info.cameraID = textBox5.Text;
            info.streamIndex = 1;
            info.capVersion = PlatformSDK.ECapabilityVersion.VAST;

            info.startTime = UtcDateTime2time_t(dateTimePickerStartTime.Value);
            info.endTime = UtcDateTime2time_t(dateTimePickerEndTime.Value);

            CreateNetFileChannel(ref info);
            CheckValidFisheyeCamera();
            var ret = PlatformSDK.SetFisheyeDisplayMode(hChannel, PlatformSDK.EFisheyeDisplayMode.mode_1O);
            EventLog.Write("PlatformSDK.SetFisheyeDisplayMode Ret = {0:X}", ret);

            
            StartSyncPlayback(ref info);
        }

        private void button2_Click(object sender, EventArgs e)
        {
            TCreateChannelInfo info = new TCreateChannelInfo();
            info.serverIP = textBox1.Text;
            info.serverUserName = textBox3.Text;
            info.serverPassword = textBox4.Text;
            info.serverPort = Convert.ToInt32(textBox2.Text);
            info.cameraID = textBox5.Text;
            info.streamIndex = 1;
            info.capVersion = PlatformSDK.ECapabilityVersion.VAST;

            CreateLiveChannel(ref info);
            CheckValidFisheyeCamera();
        }
    }
}
