﻿using System;
using System.Collections.Generic;
using System.Collections;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading;
using System.Windows.Forms;
using System.Runtime.InteropServices;
using System.Xml;
using System.IO;

namespace PlatformSDKSample
{
    using EDeviceType = PlatformSDK.EDeviceType;
    using EPTZCommand = PlatformSDK.EPTZCommand;
    using EPTZType = PlatformSDK.EPTZType;
    using EServerType = PlatformSDK.EServerType;
    using ERecordingState = PlatformSDK.ERecordingState;
    using EConnectionStatus = PlatformSDK.EConnectionStatus;
    using ECapabilityVersion = PlatformSDK.ECapabilityVersion;
    using EHardwareAcceleration = PlatformSDK.EHardwareAcceleration;

    using TBookmark = PlatformSDK.TBookmark;
    using TEventInfo = PlatformSDK.TEventInfo;
    using TAlarmInfo = PlatformSDK.TAlarmInfo;
    using TAbnormalEventInfo = PlatformSDK.TAbnormalEventInfo;
    using TLoginInfo = PlatformSDK.TLoginInfo;
    using TDeviceInfo = PlatformSDK.TDeviceInfo;
    using TDeviceDetailInfo = PlatformSDK.TDeviceDetailInfo;
    using TStreamInfo = PlatformSDK.TStreamInfo;
    using TPTZRequest = PlatformSDK.TPTZRequest;
    using TContinuousPTZRequest = PlatformSDK.TContinuousPTZRequest;
    using THttpRequest = PlatformSDK.THttpRequest;
    using TBookmarkList = PlatformSDK.TBookmarkList;
    using TRecordingDate = PlatformSDK.TRecordingDate;
    using TEventListInfo = PlatformSDK.TEventListInfo;
    using TSnapshotCBInfo = PlatformSDK.TSnapshotCallbackInfo;
    using TExportFileInfo = PlatformSDK.TExportFileInfo;
    using TManualAlarmInfo = PlatformSDK.TManualAlarmInfo;
    using TVideoDecodeInfo = PlatformSDK.TVideoDecodeInfo;
    using TBookmarkRequest = PlatformSDK.TBookmarkRequest;
    using TCameraStatusList = PlatformSDK.TCameraStatusList;
    using TDeviceStatusInfo = PlatformSDK.TDeviceStatusInfo;
    using TStreamStatusInfo = PlatformSDK.TStreamStatusInfo;
    using TServerStatusInfo = PlatformSDK.TServerStatusInfo;
    using TExportStreamInfo = PlatformSDK.TExportStreamInfo;
    using TRecordingTimeList = PlatformSDK.TRecordingTimeList;
    using TRecordingDateList = PlatformSDK.TRecordingDateList;
    using TDeviceListReqInfo = PlatformSDK.TDeviceListRequest;
    using TDeviceListResInfo = PlatformSDK.TDeviceListResponse;
    using TDeviceDetailListResponse = PlatformSDK.TDeviceDetailListResponse;
    using TAlarmNotification = PlatformSDK.TAlarmNotification;
    using TBlockedRequestInfo = PlatformSDK.TBlockedRequestInfo;
    using TChannelSnapshotInfo = PlatformSDK.TChannelSnapshotInfo;
    using TBlockedResponseInfo = PlatformSDK.TBlockedResponseInfo;
    using TRecordingDeviceList = PlatformSDK.TRecordingDeviceList;
    using TRecordingDeviceInfo = PlatformSDK.TRecordingDeviceInfo;
    using TLiveSnapshotRequest = PlatformSDK.TLiveSnapshotRequest;
    using TRecordingDateReqInfo = PlatformSDK.TRecordingDateReqInfo;
    using TNonBlockedRequestInfo = PlatformSDK.TNonBlockedRequestInfo;
    using TGetCameraStatusReqInfo = PlatformSDK.TGetCameraStatusReqInfo;
    using TPlaybackSnapshotRequest = PlatformSDK.TPlaybackSnapshotRequest;
    using TDeviceRecordingStatusInfo = PlatformSDK.TDeviceRecordingStatusInfo;
    using TSDKPresetList = PlatformSDK.TSDKPresetList;
    using TSDKPreset = PlatformSDK.TSDKPreset;
    using System.Threading.Tasks;

    public enum EFisheyeDispType
    {
        display_1O,
        display_1R,
        display_1P,
        display_2P,
        display_1P2R,
        display_1O3R,
        display_4R,
        display_1P3R,
        display_1O8R,
    }

    public partial class MainUI : Form
    {
        const string DEVICE_TYPE_NODE = "DeviceType";
        const string STATION_NODE = "Station";
        const string CAMERA_NODE = "Camera";
        const string IOBOX_NODE = "IOBox";
        const string DIDO_NODE = "DiDo";

        const int MAX_PATH = 260;
        const int MAX_SCHEDULED_CHANNEL_WAITING_SECONDS = 30;
        const int MAX_SCHEDULED_CHANNEL_DELETE_WAITING_SECONDS = 5;
        const int MILLISEC_PER_SECOND = 1000;
        const int MAX_HASH_TABLE_SIZE = 4096;
        const uint MAX_HEARTBEAT_LOST_TIMES = 3;
        const int MAX_GET_REC_DEVICE_WAIT_MILSEC = 5000;

        const int SERVER_TYPE_VAST = 0;
        const int SERVER_TYPE_CMS = 1;

        int bReloginConfigServer = 0;
        int bReloginEventServer = 0;
        int bDisableGetDeviceList = 0;

        uint configServerLostRetryTimes = 0;
        int newExportingProgress = 0;

        IntPtr syncPlaybackGroupHandle = IntPtr.Zero;
        List<IntPtr> channelHandles = new List<IntPtr>();

        GCHandle gcThis;

        public IntPtr loginID = IntPtr.Zero;    //station login handle
        ECapabilityVersion eCapVersion;
        string stationID;
        string serverTimeZone = "00:00";

        System.Timers.Timer loopUpdateUITimer = new System.Timers.Timer(MILLISEC_PER_SECOND);
        System.Timers.Timer loopGetDeviceListTimer = new System.Timers.Timer(1000 * 60 * 10);//60);//(1000 * 60);
        Thread thrdLoopGetLiveChannelSnapshot = null;
        bool keepCheckInactiveScheduledChannel = true;

        int loopTimesGetDeviceList = 0;

        public Dictionary<string, string> deviceDisplayNameMap = new Dictionary<string, string>();

        readonly object deviceStatusListLock = new Object();
        readonly object eventListLock = new Object();
        readonly object alarmListLock = new Object();
        readonly object didoNodeListLock = new Object();
        readonly object scheduledChannelTableLock = new Object();
        readonly object alarmNotificationLock = new Object();

        List<TDeviceStatusInfo> deviceStatusList;
        List<TEventInfo> eventList;
        List<AlarmInfo> alarmList;
        List<TDeviceInfo> deviceList;
        List<TEventInfo> didoNodeList;
        List<TAlarmNotificationItem> alarmNotificationList;
        List<TDeviceDetailInfo> deviceDetailList;

        Hashtable scheduledChannelTable;
        public TPTZRequest ptzRequest;
        public TContinuousPTZRequest continuousPTZRequest;

        PlatformSDK.pfMessCallBack messCB = null;
        PlatformSDK.pfAlarmNotificationCallBack alarmCB = null;
        PlatformSDK.pfEventCallBack eventCB = null;
        PlatformSDK.pfDeviceStatusCallBack deviceStatusCB = null;
        PlatformSDK.pfExportInfoCallBack exportInfoCB = null;
        PlatformSDK.pfSnapshotCallBack snapshotCB = null;
        PlatformSDK.pfVideoDecodeCallBack videoDecodeCB = null;
        PlatformSDK.pfStreamStatusCallBack streamStatusCB = null;
        PlatformSDK.pfStreamStatusCallBack exportStatusCB = null;
        PlatformSDK.pfNonBlockedRequestCallBack nonBlockedRequestCB = null;
        PlatformSDK.pfHttpRequestCallBack httpRequestCB = null;

        delegate void UpdateUIDelegate();
        UpdateUIDelegate updateUIDelegate = null;

        delegate void UpdateAlarmWindowDelegate(string deviceID, string status, string time, string args);
        UpdateAlarmWindowDelegate updateAlarmWindowDelegate = null;

        TDeviceInfo curSelectedDevice;      //the current selected device in device tree view
        List<TDeviceInfo> curCheckedDevices = new List<TDeviceInfo>();
        TDeviceInfo dummyDevice;

        FileStream logFile;
        StreamWriter logWriter;

        PTZPanel ptzPanel;
        RecordingDataPanel recordingDataPanel;
        RecordingGroupPanel recordingGroupPanel;

        int serverType = SERVER_TYPE_VAST;
        List<ViewCell> viewCells = new List<ViewCell>();

        enum ETunnelState : int
        {
            offline,           //when receive eDisconnected, we then change state to eOffLine
            online,            //when receive eTunnelConnect, we then change state to eOnLine
            connected,         //mark on tunnel establish
            disconnected,      //mark on tunnel disconnect            
        }

        int configServerState = (int)ETunnelState.offline;
        int eventServerState = (int)ETunnelState.offline;
        int queryServerState = (int)ETunnelState.offline;

        public struct TSnapshotContext
        {
            public string savePath;
        }

        struct TAlarmNotificationItem
        {
            public string device;
            public string state;
            public string time;
        }

        public struct AlarmInfo
        {
            public int id;
            public string enabled;
            public string alarmName;
            public string content;
            public string deviceID;
            public string DOIndex;
            public List<string> cameraIDList;
        }

        class ScheduledChannelInfo
        {
            public IntPtr hChannel;
            public EChannelMode mode;
            public IntPtr context;
            public int waitingSeconds = MAX_SCHEDULED_CHANNEL_WAITING_SECONDS;
            public int deleteWaitingSeconds = MAX_SCHEDULED_CHANNEL_DELETE_WAITING_SECONDS;
            public bool canBeClose = false;
        }

        void InitializeUI()
        {
            tabControl.SelectedIndexChanged += tabControl_SelectedIndexChanged;

            gcThis = GCHandle.Alloc(this);

            ptzPanel = new PTZPanel(this);
            recordingDataPanel = new RecordingDataPanel(this);
            recordingGroupPanel = new RecordingGroupPanel(this);

            dummyDevice = new TDeviceInfo();
            dummyDevice.deviceType = EDeviceType.unknown;
            curSelectedDevice = dummyDevice;

            deviceStatusList = new List<TDeviceStatusInfo>();
            eventList = new List<TEventInfo>();
            alarmList = new List<AlarmInfo>();
            deviceList = new List<TDeviceInfo>();
            didoNodeList = new List<TEventInfo>();
            alarmNotificationList = new List<TAlarmNotificationItem>();
            deviceDetailList = new List<TDeviceDetailInfo>();

            scheduledChannelTable = new Hashtable(MAX_HASH_TABLE_SIZE);

            messCB = new PlatformSDK.pfMessCallBack(LoginCallBack);
            alarmCB = new PlatformSDK.pfAlarmNotificationCallBack(AlarmCallBack);
            eventCB = new PlatformSDK.pfEventCallBack(EventCallBack);
            deviceStatusCB = new PlatformSDK.pfDeviceStatusCallBack(DeviceStatusCallBack);
            exportStatusCB = new PlatformSDK.pfStreamStatusCallBack(ExportFinishCallBack);
            exportInfoCB = new PlatformSDK.pfExportInfoCallBack(ExportInfoCallBack);
            snapshotCB = new PlatformSDK.pfSnapshotCallBack(SnapshotCallBack);
            videoDecodeCB = new PlatformSDK.pfVideoDecodeCallBack(VideoDecodeCallBack);
            streamStatusCB = new PlatformSDK.pfStreamStatusCallBack(StreamStatusCallBack);
            nonBlockedRequestCB = new PlatformSDK.pfNonBlockedRequestCallBack(NonBlockedRequestCallBack);
            httpRequestCB = new PlatformSDK.pfHttpRequestCallBack(HttpRequestCallBack);

            updateUIDelegate = new UpdateUIDelegate(UpdateUI);
            updateAlarmWindowDelegate = new UpdateAlarmWindowDelegate(UpdateAlarmWindow);

            loopGetDeviceListTimer.Elapsed += new System.Timers.ElapsedEventHandler(LoopGetDeviceList);

            loopUpdateUITimer.Elapsed += new System.Timers.ElapsedEventHandler(LoopUpdateUI);
            loopUpdateUITimer.AutoReset = true;
            loopUpdateUITimer.Enabled = false;

            thrdLoopGetLiveChannelSnapshot = new Thread(LoopReleaseInactiveScheduledChannel);
            thrdLoopGetLiveChannelSnapshot.Start();

            eventWindow.Columns.Add("Device", 340, HorizontalAlignment.Left);
            eventWindow.Columns.Add("Event type", 120, HorizontalAlignment.Left);
            eventWindow.Columns.Add("Time", 180, HorizontalAlignment.Left);

            alarmWindow.Columns.Add("Device", 340, HorizontalAlignment.Left);
            alarmWindow.Columns.Add("Alarm type", 120, HorizontalAlignment.Left);
            alarmWindow.Columns.Add("Time", 180, HorizontalAlignment.Left);

            comboBoxServerType.SelectedIndex = 0;
        }

        private void InitializeLog()
        {
            String logFileName = @".\log.txt";
            try
            {
                logFile = new FileStream(logFileName, FileMode.OpenOrCreate, FileAccess.ReadWrite);
                logWriter = new StreamWriter(logFile);
            }
            catch (IOException e)
            {
                MessageBox.Show("Fail to open " + logFileName + ": " + e.ToString());
            }
        }

        private void UninitializeLog()
        {
            logWriter.Close();
            logFile.Close();
        }

        public MainUI()
        {
            InitializeComponent();
            InitializeUI();
            InitializeLog();

            int ret = PlatformSDK.Initial();
            EventLog.Write("PlatformSDK.Initial Ret = {0:X}", ret);

            //TestLoopLoginout();
        }

        ~MainUI()
        {
            int ret = PlatformSDK.Release();
            EventLog.Write("PlatformSDK.Release Ret = {0:X}", ret);

            UninitializeLog();

            gcThis.Free();
        }

        void HandleAlarm(string deviceID, string status, string time, string args)
        {
            Invoke(updateAlarmWindowDelegate, deviceID, status, time, args);
        }

        string GetDeviceDisplayName(string deviceID)
        {
            if (deviceDisplayNameMap.ContainsKey(deviceID))
            {
                return deviceDisplayNameMap[deviceID];
            }

            return "";
        }

        void UpdateAlarmWindow(string deviceID, string status, string time, string args)
        {
            ListViewItem lvi = new ListViewItem();

            lvi.Text = GetDeviceDisplayName(deviceID);
            lvi.SubItems.Add(status);
            lvi.SubItems.Add(time);

            alarmWindow.Items.Add(lvi);
        }

        void HandleConfigServerStatusChange(PlatformSDK.EConnectionStatus status)
        {
            switch (status)
            {
                case PlatformSDK.EConnectionStatus.connected:
                    Interlocked.Exchange(ref configServerState, (int)ETunnelState.connected);
                    loopUpdateUITimer.Enabled = true;
                    break;

                case PlatformSDK.EConnectionStatus.disconnected:
                    Interlocked.Exchange(ref configServerState, (int)ETunnelState.disconnected);
                    break;

                case PlatformSDK.EConnectionStatus.receiveHeartBeat:
                    configServerLostRetryTimes = 0;
                    break;

                case PlatformSDK.EConnectionStatus.heartBeatLostRetry:
                    configServerLostRetryTimes++;
                    logWriter.WriteLine("heartbeat lost retry time = " + configServerLostRetryTimes);
                    break;

                case PlatformSDK.EConnectionStatus.heartBeatLost:
                    logWriter.WriteLine("heartbeat lost");
                    Interlocked.Exchange(ref configServerState, (int)ETunnelState.disconnected);
                    break;
            }
        }

        void HandleEventServerStatusChange(PlatformSDK.EConnectionStatus status)
        {
            switch (status)
            {
                case PlatformSDK.EConnectionStatus.connected:
                    Interlocked.Exchange(ref eventServerState, (int)ETunnelState.connected);
                    break;

                case PlatformSDK.EConnectionStatus.disconnected:
                    Interlocked.Exchange(ref eventServerState, (int)ETunnelState.disconnected);
                    break;
            }
        }

        void HandleQueryServerStatusChange(PlatformSDK.EConnectionStatus status)
        {
            switch (status)
            {
                case PlatformSDK.EConnectionStatus.connected:
                    Interlocked.Exchange(ref queryServerState, (int)ETunnelState.connected);
                    break;

                case PlatformSDK.EConnectionStatus.disconnected:
                    Interlocked.Exchange(ref queryServerState, (int)ETunnelState.disconnected);
                    break;
            }
        }

        string GetServerTypeString(EServerType type)
        {
            if (type == EServerType.configServer) return "config server";
            if (type == EServerType.eventServer) return "event server";
            if (type == EServerType.queryServer) return "query server";
            return "unknown server";
        }

        string GetConnStatusString(PlatformSDK.EConnectionStatus status)
        {
            switch (status)
            {
                case EConnectionStatus.connected: return "connected";
                case EConnectionStatus.disconnected: return "disconnected";
                case EConnectionStatus.receiveHeartBeat: return "receive heartbeat";
                case EConnectionStatus.heartBeatLost: return "heartbeat lost";
                case EConnectionStatus.heartBeatLostRetry: return "heartbeat lost retry, times = " + configServerLostRetryTimes;
                case EConnectionStatus.unchanged: return "unchanged";
            }
            return "unknown";
        }

        string GetRecordingStatusString(PlatformSDK.ERecordingState status)
        {
            switch (status)
            {
                case ERecordingState.none: return "none";
                case ERecordingState.scheduled: return "scheduled";
                case ERecordingState.manual: return "manual";
                case ERecordingState.byEvent: return "event";
                case ERecordingState.continuous: return "continuous";
            }
            return "unknown";
        }

        void LoginCallBack(IntPtr pStatusInfo)
        {
            TServerStatusInfo info = (TServerStatusInfo)Marshal.PtrToStructure(pStatusInfo, typeof(TServerStatusInfo));

            if (info.status != PlatformSDK.EConnectionStatus.receiveHeartBeat)
            {
                EventLog.Write("LoginCallBack: " + info.hLoginID
                    + ", " + GetServerTypeString(info.serverType)
                    + " " + GetConnStatusString(info.status));
            }

            if (info.serverType == EServerType.configServer)
            {
                HandleConfigServerStatusChange(info.status);
            }
            else if (info.serverType == EServerType.eventServer)
            {
                HandleEventServerStatusChange(info.status);
            }
            else if (info.serverType == EServerType.queryServer)
            {
                HandleQueryServerStatusChange(info.status);
            }
        }

        void ParseAlarm(TAlarmNotification alarmInfo)
        {
            string msg = Marshal.PtrToStringAnsi(alarmInfo.content);

            XmlDocument doc = new XmlDocument();

            doc.LoadXml(msg);

            XmlNodeList deviceID;
            XmlNodeList state;

            if (doc.GetElementsByTagName("CamEvent").Count > 0)
            {
                state = doc.GetElementsByTagName("EventType");
                deviceID = doc.GetElementsByTagName("CamName");
            }
            else if (doc.GetElementsByTagName("CamConn").Count > 0)
            {
                state = doc.GetElementsByTagName("State");
                deviceID = doc.GetElementsByTagName("CamName");
            }
            else if (doc.GetElementsByTagName("RecStatus").Count > 0)
            {
                state = doc.GetElementsByTagName("State");
                deviceID = doc.GetElementsByTagName("CamName");
            }
            else if (doc.GetElementsByTagName("RecErr").Count > 0)
            {
                state = doc.GetElementsByTagName("State");
                deviceID = doc.GetElementsByTagName("CamName");
            }
            else if (doc.GetElementsByTagName("StationConn").Count > 0)
            {
                state = doc.GetElementsByTagName("State");
                deviceID = doc.GetElementsByTagName("StationAddress");
            }
            else if (doc.GetElementsByTagName("StorageConn").Count > 0)
            {
                state = doc.GetElementsByTagName("State");
                deviceID = doc.GetElementsByTagName("StationAddress");
            }
            else if (doc.GetElementsByTagName("StorageCapacity").Count > 0)
            {
                state = doc.GetElementsByTagName("State");
                deviceID = doc.GetElementsByTagName("StationAddress");
            }
            else if (doc.GetElementsByTagName("DIOStatus").Count > 0)
            {
                state = doc.GetElementsByTagName("EventType");
                deviceID = doc.GetElementsByTagName("DeviceName");
            }
            else if (doc.GetElementsByTagName("SysLicense").Count > 0)
            {
                state = doc.GetElementsByTagName("State");
                deviceID = doc.GetElementsByTagName("StationAdDress");
            }
            else if (doc.GetElementsByTagName("SysNetwork").Count > 0)
            {
                state = doc.GetElementsByTagName("State");
                deviceID = doc.GetElementsByTagName("StationAdDress");
            }
            else if (doc.GetElementsByTagName("SysVirtualMem").Count > 0)
            {
                state = doc.GetElementsByTagName("State");
                deviceID = doc.GetElementsByTagName("StationAdDress");
            }
            else if (doc.GetElementsByTagName("ExtIoDevice").Count > 0)
            {
                state = doc.GetElementsByTagName("EventType");
                deviceID = doc.GetElementsByTagName("DeviceName");
            }
            else if (doc.GetElementsByTagName("ManualEvent").Count > 0)
            {
                state = doc.GetElementsByTagName("Description");
                deviceID = doc.GetElementsByTagName("StationAdDress");
            }
            else
            {
                return;
            }

            bool bAlarmNotification = false;

            if (deviceID.Count == 0 || state.Count == 0)
            {
                if (doc.GetElementsByTagName("AlarmNotification").Count == 0) return;
                state = doc.GetElementsByTagName("EventName");
                if (state.Count == 0) return;
                bAlarmNotification = true;
            }

            XmlNodeList time = doc.GetElementsByTagName("ServerTime");
            if (time.Count == 0) return;

            TAlarmNotificationItem alarm = new TAlarmNotificationItem();

            alarm.device = bAlarmNotification ? "AlarmNotification" : deviceID[0].InnerText;
            alarm.state = state[0].InnerText;
            alarm.time = time[0].InnerText;

            lock (alarmNotificationLock)
            {
                alarmNotificationList.Add(alarm);
            }
        }

        void UpdateAlarmNotification()
        {
            lock (alarmNotificationLock)
            {
                for (int i = alarmNotificationList.Count - 1; i >= 0; --i)
                {
                    UpdateAlarmWindow(alarmNotificationList[i].device
                        , alarmNotificationList[i].state
                        , alarmNotificationList[i].time
                        , null);
                }
                alarmNotificationList.Clear();
            }
        }

        void AlarmCallBack(IntPtr pAlarmList, int alarmNum, IntPtr context)
        {
            EventLog.Write("Receive alarm num = " + alarmNum);

            for (int i = 0; i < alarmNum; ++i)
            {
                IntPtr pAlarmInfo = new IntPtr(pAlarmList.ToInt64() + Marshal.SizeOf(typeof(TAlarmNotification)) * i);
                TAlarmNotification alarmInfo = (TAlarmNotification)Marshal.PtrToStructure(pAlarmInfo, typeof(TAlarmNotification));
                string str = Marshal.PtrToStringAnsi(alarmInfo.content);
                EventLog.Write("len = " + alarmInfo.contentLen
                    + ", context = " + str
                );

                ParseAlarm(alarmInfo);
            }
        }

        unsafe void SnapshotCallBack(IntPtr pSnapshotInfo)
        {
            TSnapshotCBInfo snapshotInfo = (TSnapshotCBInfo)Marshal.PtrToStructure(pSnapshotInfo, typeof(TSnapshotCBInfo));
            TSnapshotContext context = (TSnapshotContext)Marshal.PtrToStructure(snapshotInfo.context, typeof(TSnapshotContext));

            if (snapshotInfo.imageSize == 0)
            {
                EventLog.Write("Fail to get snapshot: size = 0");
                return;
            }

            byte[] image = new byte[snapshotInfo.imageSize];
            Marshal.Copy(snapshotInfo.image, image, 0, snapshotInfo.imageSize);

            try
            {
                BinaryWriter writer = new BinaryWriter(File.Open(context.savePath, FileMode.OpenOrCreate));
                writer.Write(image, 0, snapshotInfo.imageSize);
                writer.Flush();
                writer.Close();
            }
            catch (IOException e)
            {
                EventLog.Write("Fail to write file \"" + context.savePath + "\":\n" + e.ToString());
            }
        }

        void ExportInfoCallBack(IntPtr pExportInfo)
        {
            TExportStreamInfo info = (TExportStreamInfo)Marshal.PtrToStructure(pExportInfo, typeof(TExportStreamInfo));
            string str = Marshal.PtrToStringAnsi(info.context);

            newExportingProgress = info.exportingProgress;
        }

        unsafe void ExportFinishCallBack(IntPtr pStatusInfo)
        {
            TStreamStatusInfo info = (TStreamStatusInfo)Marshal.PtrToStructure(pStatusInfo, typeof(TStreamStatusInfo));
            string str = Marshal.PtrToStringAnsi(info.context);
            EventLog.Write("StreamStatusCallBack: hChannel = {0:X}, Status = {1}, {2}", info.hChannel, info.status, str);

            newExportingProgress = 100;

            lock (scheduledChannelTableLock)
            {
                if (!scheduledChannelTable.Contains(info.hChannel)) return;
                ((ScheduledChannelInfo)(scheduledChannelTable[info.hChannel])).canBeClose = true;
            }
        }

        void EventCallBack(IntPtr pEventListInfo)
        {
            TEventListInfo info = (TEventListInfo)Marshal.PtrToStructure(pEventListInfo, typeof(TEventListInfo));

            for (uint i = 0; i < info.eventNum; ++i)
            {
                IntPtr pEventInfo = new IntPtr(info.pEventList.ToInt64() + Marshal.SizeOf(typeof(TEventInfo)) * i);
                TEventInfo eventInfo = (TEventInfo)Marshal.PtrToStructure(pEventInfo, typeof(TEventInfo));
                AddEventInfo(ref eventInfo);
            }
        }

        uint image_idx = 1;

        unsafe void SaveImage(ref TVideoDecodeInfo decodeInfo)
        {
            if (decodeInfo.pBuffer == IntPtr.Zero) return;

            //string savePath = textBoxDir.Text + "\\" + textBoxFilename.Text;
            System.DateTime time = new System.DateTime(1970, 1, 1).AddSeconds(decodeInfo.frameSecond);
            string savePath = textBoxDir.Text + "\\" + time.ToString("yyyyMMdd_HHmmss")
                + "_" + decodeInfo.frameMilSec + "(" + decodeInfo.hChannel + ") [" + image_idx + "].jpg";
            image_idx++;

            if (System.IO.File.Exists(savePath))
            {
                EventLog.Write("File already exists: [{0:X}] size = {1}, name = {2}"
                    , decodeInfo.hChannel
                    , decodeInfo.bufSize
                    , savePath);
                return;
            }
            else
            {
                EventLog.Write("Save image: [{0:X}] size = {1}, name = {2}"
                    , decodeInfo.hChannel
                    , decodeInfo.bufSize
                    , savePath);
            }

            try
            {
                byte[] image = new byte[decodeInfo.bufSize];
                Marshal.Copy(decodeInfo.pBuffer, image, 0, decodeInfo.bufSize);
                FileStream file = new FileStream(savePath, FileMode.OpenOrCreate, FileAccess.Write);
                file.Write(image, 0, decodeInfo.bufSize);
                file.Close();
            }
            catch (IOException e)
            {
                EventLog.Write("Fail to save image \"" + savePath + "\":\n" + e.ToString());
            }
        }

        void VideoDecodeCallBack(IntPtr pVideoDecodeInfo)
        {
            TVideoDecodeInfo decodeInfo = (TVideoDecodeInfo)Marshal.PtrToStructure(pVideoDecodeInfo, typeof(TVideoDecodeInfo));

            if (decodeInfo.pBuffer == IntPtr.Zero) return;

            ScheduledChannelInfo chInfo = null;

            lock (scheduledChannelTableLock)
            {
                if (!scheduledChannelTable.Contains(decodeInfo.hChannel)) return;
                chInfo = (ScheduledChannelInfo)scheduledChannelTable[decodeInfo.hChannel];
                if (chInfo.canBeClose) return;
            }

            SaveImage(ref decodeInfo);

            lock (scheduledChannelTableLock)
            {
                chInfo.canBeClose = true;
            }
        }

        void StreamStatusCallBack(IntPtr pStatusInfo)
        {
            StreamStatusHandler(pStatusInfo);
        }

        void AddDeviceStatusInfo(ref TDeviceStatusInfo info, IntPtr context)
        {
            lock (deviceStatusListLock)
            {
                deviceStatusList.Add(info);
            }
        }

        void AddEventInfo(ref TEventInfo info)
        {
            lock (eventListLock)
            {
                eventList.Add(info);
            }
        }

        void DeviceStatusCallBack(IntPtr pStatusList, int statusNum, IntPtr context)
        {
            for (int i = 0; i < statusNum; ++i)
            {
                IntPtr pEventInfo = new IntPtr(pStatusList.ToInt64() + Marshal.SizeOf(typeof(TDeviceStatusInfo)) * i);
                TDeviceStatusInfo statusInfo = (TDeviceStatusInfo)Marshal.PtrToStructure(pEventInfo, typeof(TDeviceStatusInfo));
                AddDeviceStatusInfo(ref statusInfo, context);
            }
        }

        void CheckCameraStatus()
        {
            TCameraStatusList camStatusList = new TCameraStatusList();
            TGetCameraStatusReqInfo request = new TGetCameraStatusReqInfo();

            request.getAllCameraStatus = true;
            request.timeoutMillisecond = 5000;

            int ret = PlatformSDK.GetCameraStatus(loginID, ref request, ref camStatusList);
            EventLog.Write("PlatformSDK.GetCameraStatus Ret = {0:X}", ret);
            if (ret != 0) return;

            for (uint i = 0; i < camStatusList.statusNum; ++i)
            {
                IntPtr pStatusInfo = new IntPtr(camStatusList.statusList.ToInt64() + Marshal.SizeOf(typeof(TDeviceStatusInfo)) * i);
                TDeviceStatusInfo status = (TDeviceStatusInfo)Marshal.PtrToStructure(pStatusInfo, typeof(TDeviceStatusInfo));

                string info = status.deviceID + " (" + GetConnStatusString(status.connectionStatus) + ")";
                if (status.recordingStatusInfo != IntPtr.Zero)
                {
                    IntPtr pRecStatus = new IntPtr(status.recordingStatusInfo.ToInt64());
                    TDeviceRecordingStatusInfo recStatus =
                        (TDeviceRecordingStatusInfo)Marshal.PtrToStructure(pRecStatus, typeof(TDeviceRecordingStatusInfo));
                    info += ", recording state: " + GetRecordingStatusString(recStatus.deviceRecordingState);
                    if (recStatus.groupID > 0)
                    {
                        info += ", Group ID: " + recStatus.groupID.ToString();
                        info += " (" + GetRecordingStatusString(recStatus.groupRecordingState) + ")";
                    }
                }

                EventLog.Write(info);
            }
        }

        void MonitorEvent(IntPtr loginID)
        {
            int ret = PlatformSDK.MonitorEvent(loginID, eventCB, IntPtr.Zero);
            EventLog.Write("PlatformSDK.MonitorEvent Ret = {0:X}", ret);
        }

        unsafe private void MonitorDeviceStatus(IntPtr loginID)
        {
            PlatformSDK.MonitorDeviceStatus(loginID, deviceStatusCB, IntPtr.Zero);
        }

        void ParseAlarmInfo(TAlarmInfo rawAlarmInfo, ref AlarmInfo alarm)
        {
            alarm.id = rawAlarmInfo.ID;
            alarm.alarmName = rawAlarmInfo.alarmName;
            alarm.content = Marshal.PtrToStringAnsi(rawAlarmInfo.content);

            XmlDocument doc = new XmlDocument();

            doc.LoadXml(alarm.content);

            XmlNodeList enabled = doc.GetElementsByTagName("Enabled");
            if (enabled.Count == 1)
            {
                alarm.enabled = enabled[0].InnerText;
            }

            XmlNodeList deviceID = doc.GetElementsByTagName("DeviceName");
            if (deviceID.Count == 1)
            {
                alarm.deviceID = deviceID[0].InnerText;
            }

            XmlNodeList DOIndex = doc.GetElementsByTagName("Index");
            if (DOIndex.Count == 1)
            {
                alarm.DOIndex = DOIndex[0].InnerText;
            }

            alarm.cameraIDList = new List<string>();
            XmlNodeList CamIDList = doc.GetElementsByTagName("Cam");
            for (int i = 0; i < CamIDList.Count; ++i)
            {
                alarm.cameraIDList.Add(CamIDList[i].InnerText);
            }
        }

        unsafe private void GetAlarmList(List<AlarmInfo> listAlarm)
        {
            listAlarm.Clear();

            int alarmNum = 0;
            IntPtr pAlarmList = IntPtr.Zero;
            int ret = PlatformSDK.GetAlarmInfoList(loginID, ref pAlarmList, &alarmNum);
            EventLog.Write("PlatformSDK.GetAlarmInfoList Ret = {0:X}", ret);

            for (int i = 0; i < alarmNum; ++i)
            {
                IntPtr pAlarmInfo = new IntPtr(pAlarmList.ToInt64() + Marshal.SizeOf(typeof(TAlarmInfo)) * i);
                TAlarmInfo rawAlarmInfo = (TAlarmInfo)Marshal.PtrToStructure(pAlarmInfo, typeof(TAlarmInfo));
                AlarmInfo alarm = new AlarmInfo();

                ParseAlarmInfo(rawAlarmInfo, ref alarm);
                listAlarm.Add(alarm);
            }
        }

        unsafe private void TestTriggerAlarm(List<TAlarmInfo> listAlarm)
        {
            foreach (TAlarmInfo alarm in listAlarm)
            {
                string str = Marshal.PtrToStringAnsi(alarm.content);
                EventLog.Write("ID = " + alarm.ID
                    + ", name = " + alarm.alarmName
                    + ", len = " + alarm.contentLen
                    + ", context = " + str);

                TManualAlarmInfo alarmInfo = new TManualAlarmInfo();
                alarmInfo.ID = alarmList[comboBoxAlarm.SelectedIndex].id;

                string description = textBoxAlarmDescription.Text;
                if (!string.IsNullOrEmpty(description))
                {
                    alarmInfo.szUTF8Description = new byte[PlatformSDK.MAX_ALARM_DESC_LEN + 1];
                    byte[] asciiDescription = Encoding.UTF8.GetBytes(description);
                    Array.Copy(asciiDescription, alarmInfo.szUTF8Description, asciiDescription.Length);
                }

                int ret = PlatformSDK.TriggerAlarm(loginID, ref alarmInfo);
                EventLog.Write("PlatformSDK.TriggerAlarm Ret = {0:X}", ret);
            }
        }

        string GetServerVersion(ref PlatformSDK.TServerInfo serverInfo)
        {
            string capVersion = "Unknown ";
            switch (serverInfo.capVersion)
            {
                case ECapabilityVersion.VAST:
                    capVersion = "VAST ";
                    break;
                case ECapabilityVersion.NVR:
                    capVersion = "NVR ";
                    break;
                case ECapabilityVersion.CMS:
                    capVersion = "CMS ";
                    break;
            }
            return capVersion + serverInfo.version;
        }

        unsafe private void GetServerInfo()
        {
            PlatformSDK.TServerInfo serverInfo = new PlatformSDK.TServerInfo();
            int ret = PlatformSDK.GetServerInfo(loginID, ref serverInfo);
            EventLog.Write("PlatformSDK.GetServerInfo Ret = {0:X}", ret);
            if (ret != 0) return;
            labelServerVersionInfo.Text = GetServerVersion(ref serverInfo);
            eCapVersion = serverInfo.capVersion;
            stationID = serverInfo.stationID;
        }

        unsafe private void MonitorAlarmNotification()
        {
            int ret = PlatformSDK.MonitorAlarmNotification(loginID, alarmCB, (IntPtr)gcThis);
            EventLog.Write("PlatformSDK.MonitorAlarmNotification Ret = " + ret);
        }

        private void LoopUpdateExportingProgress(object source, System.Timers.ElapsedEventArgs e)
        {
            //labelExportProgress.Text = "" + newExportingProgress.ToString() + " %";
        }

        unsafe private void LoopGetDeviceList(object source, System.Timers.ElapsedEventArgs e)
        {
            if (serverType == SERVER_TYPE_VAST)
            {
                TestGetDeviceList(loginID);
            }
        }

        void HandleConfigServerDisconnect()
        {
            Interlocked.Exchange(ref configServerState, (int)ETunnelState.offline);
        }

        void HandleEventServerDisconnect()
        {
            Interlocked.Exchange(ref eventServerState, (int)ETunnelState.offline);
        }

        void HandleQueryServerDisconnect()
        {
            Interlocked.Exchange(ref queryServerState, (int)ETunnelState.offline);
        }

        void HandleConfigServerConnect()
        {
            Interlocked.Exchange(ref configServerState, (int)ETunnelState.online);

            if (bDisableGetDeviceList == 1)
            {
                Interlocked.Exchange(ref bDisableGetDeviceList, 0);
            }

            GetServerInfo();

            GetServerTimeZone();

            GetDeviceListIntoDeviceTree();

            GetDeviceListWithDetail();

            //@For test
            //TestLoopGetLiveChannelSnapshot();
            //TestExportFile(loginID);


            LoopGetDeviceList(loginID);

            CheckCameraStatus();
        }

        void HandleEventServerConnect()
        {
            Interlocked.Exchange(ref eventServerState, (int)ETunnelState.online);

            GetAlarmListIntoComboAlarm();

            MonitorEvent(loginID);
            MonitorAlarmNotification();
            MonitorDeviceStatus(loginID);
        }

        void HandleEventServerOnLine()
        {
            UpdateAlarmNotification();
        }

        void HandleQueryServerConnect()
        {
            Interlocked.Exchange(ref queryServerState, (int)ETunnelState.online);

            //GetRecordingDeviceList();   //@sam
        }

        void HandleQueryServerOnLine()
        {
        }

        void HandleConfigServerOnLine()
        {
            UpdateEventWindowAndDeviceTree();
            UpdateDeviceStatus();
        }

        void StreamStatusHandler(IntPtr pStatusInfo)
        {
            TStreamStatusInfo info = (TStreamStatusInfo)Marshal.PtrToStructure(pStatusInfo, typeof(TStreamStatusInfo));
            EventLog.Write(DateTime.Now.ToString("yyyy/MM/dd HH:mm:ss.fff") +
                "  StreamStatusCallBack: hChannel = {0:X}, Status = {1}", info.hChannel, info.status);

            lock (scheduledChannelTableLock)
            {
                if (!scheduledChannelTable.Contains(info.hChannel)) return;
                ScheduledChannelInfo chInfo = (ScheduledChannelInfo)scheduledChannelTable[info.hChannel];
                if (chInfo.canBeClose) return;
                if (info.status == PlatformSDK.EStreamStatus.stopped)
                {
                    chInfo.canBeClose = true;
                }
            }
        }

        void UpdateUI()
        {
            switch ((ETunnelState)configServerState)
            {
                case ETunnelState.connected:
                    HandleConfigServerConnect();
                    break;
                case ETunnelState.disconnected:
                    HandleConfigServerDisconnect();
                    if (bReloginConfigServer == 0)
                    {
                        Logout();
                    }
                    break;
                case ETunnelState.online:
                    HandleConfigServerOnLine();
                    break;
                case ETunnelState.offline:
                    break;
            }

            switch ((ETunnelState)eventServerState)
            {
                case ETunnelState.connected:
                    HandleEventServerConnect();
                    break;
                case ETunnelState.disconnected:
                    HandleEventServerDisconnect();
                    break;
                case ETunnelState.online:
                    HandleEventServerOnLine();
                    break;
                case ETunnelState.offline:
                    break;
            }

            switch ((ETunnelState)queryServerState)
            {
                case ETunnelState.connected:
                    HandleQueryServerConnect();
                    break;
                case ETunnelState.disconnected:
                    HandleQueryServerDisconnect();
                    break;
                case ETunnelState.online:
                    HandleQueryServerOnLine();
                    break;
                case ETunnelState.offline:
                    break;
            }

            labelExportProgress.Text = newExportingProgress.ToString() + " %";
        }

        void LoopUpdateUI(object source, System.Timers.ElapsedEventArgs e)
        {
            if (InvokeRequired)
            {
                BeginInvoke(updateUIDelegate);
                return;
            }
            else
            {
                UpdateUI();
            }
        }

        void GetRecordingDeviceList()
        {
            TDeviceListReqInfo request = new TDeviceListReqInfo();
            TRecordingDeviceList deviceList = new TRecordingDeviceList();

            request.stationID = null;
            request.timeoutMillisecond = MAX_GET_REC_DEVICE_WAIT_MILSEC;

            int ret = PlatformSDK.GetRecordingDeviceList(loginID, ref request, ref deviceList);
            EventLog.Write("PlatformSDK.GetRecordingDeviceList Ret = {0:X}", ret);

            recordingDataPanel.SetRecordingDeviceList(ref deviceList);
        }

        void CloseSnapshotChannel(ScheduledChannelInfo info)
        {
            IntPtr hOldChannel = info.hChannel;

            if (info.mode == EChannelMode.live)
            {
                int ret = PlatformSDK.DeleteLiveChannel(ref info.hChannel);
                EventLog.Write(DateTime.Now.ToString("yyyy/MM/dd HH:mm:ss.fff") +
                    "  DeleteLiveChannel {0:X}, Ret = {1:X}", hOldChannel, ret);
            }
            else if (info.mode == EChannelMode.playback)
            {
                int ret = PlatformSDK.DeleteNetFileChannel(ref info.hChannel);
                EventLog.Write(DateTime.Now.ToString("yyyy/MM/dd HH:mm:ss.fff") +
                    "  DeleteNetFileChannel {0:X}, Ret = {1:X}", hOldChannel, ret);
            }
            else if (info.mode == EChannelMode.export)
            {
                Marshal.FreeHGlobal(info.context);
                int ret = PlatformSDK.DeleteExportMediaFileChannel(ref info.hChannel);
                EventLog.Write(DateTime.Now.ToString("yyyy/MM/dd HH:mm:ss.fff") +
                    "  DeleteExportMediaFileChannel {0:X}, Ret = {1:X}", hOldChannel, ret);
            }
        }

        void CloseAllSnapshotChannel()
        {
            List<ScheduledChannelInfo> channelSnapshotRemoveList = new List<ScheduledChannelInfo>();
            List<int> hashKeyRemoveList = new List<int>();

            lock (scheduledChannelTableLock)
            {
                foreach (DictionaryEntry item in scheduledChannelTable)
                {
                    ScheduledChannelInfo info = (ScheduledChannelInfo)item.Value;
                    channelSnapshotRemoveList.Add(info);
                    hashKeyRemoveList.Add((int)item.Key);
                }
            }

            foreach (ScheduledChannelInfo info in channelSnapshotRemoveList)
            {
                CloseSnapshotChannel(info);
            }

            lock (scheduledChannelTableLock)
            {
                foreach (int key in hashKeyRemoveList)
                {
                    scheduledChannelTable.Remove(key);
                }
            }
            channelSnapshotRemoveList.Clear();
        }

        void ReleaseInactiveScheduledChannel()
        {
            List<ScheduledChannelInfo> scheduledChannelRemoveList = new List<ScheduledChannelInfo>();
            List<IntPtr> hashKeyRemoveList = new List<IntPtr>();

            lock (scheduledChannelTableLock)
            {
                foreach (DictionaryEntry item in scheduledChannelTable)
                {
                    ScheduledChannelInfo info = (ScheduledChannelInfo)item.Value;

                    if (info.canBeClose == true)
                    {
                        info.waitingSeconds = 0;
                    }

                    if (info.waitingSeconds > 0)
                    {
                        info.waitingSeconds--;
                        continue;
                    }

                    if (info.deleteWaitingSeconds > 0)
                    {
                        info.deleteWaitingSeconds--;
                        continue;
                    }

                    hashKeyRemoveList.Add((IntPtr)item.Key);
                    scheduledChannelRemoveList.Add(info);
                }
            }

            foreach (ScheduledChannelInfo info in scheduledChannelRemoveList)
            {
                CloseSnapshotChannel(info);
            }

            lock (scheduledChannelTableLock)
            {
                foreach (IntPtr key in hashKeyRemoveList)
                {
                    scheduledChannelTable.Remove(key);
                }
            }

            scheduledChannelRemoveList.Clear();
        }

        void LoopReleaseInactiveScheduledChannel()
        {
            while (keepCheckInactiveScheduledChannel)
            {
                ReleaseInactiveScheduledChannel();
                Thread.Sleep(MILLISEC_PER_SECOND);
            }
        }

        unsafe private void LoopGetDeviceList(IntPtr loginID)
        {
            loopGetDeviceListTimer.AutoReset = true;    //loop run
            loopGetDeviceListTimer.Enabled = true;
        }

        string GetDeviceTypeString(EDeviceType type)
        {
            switch (type)
            {
                case EDeviceType.camera: return "camera";
                case EDeviceType.externalDevice: return "external device";
                case EDeviceType.station: return "station";
            }
            return "unknown";
        }

        unsafe private void LogDeviceList()
        {
            if (logWriter == null) return;

            logWriter.WriteLine("-----------------------------------------------");
            logWriter.WriteLine(DateTime.Now.ToString("yyyy/MM/dd HH:mm:ss"));

            foreach (TDeviceInfo device in deviceList)
            {
                logWriter.WriteLine(GetDeviceTypeString(device.deviceType) + " " + device.IP);
            }
            logWriter.Flush();
        }

        unsafe void TestGetDeviceList(IntPtr loginID)
        {
            loopTimesGetDeviceList++;

            GetDeviceList(deviceList);

            //LogDeviceList();
        }

        private string getReferenceNameSafely(in string refName)
        {
            return refName.Length < 2 ? "" : refName.Substring(2);
        }

        string GetTreeNodeName(TDeviceInfo device)
        {
            string name = device.deviceID + ", " + getReferenceNameSafely(device.refName)+ " (" + device.IP + ")";
            if (device.deviceType == EDeviceType.camera)
            {
                name += ", ch = " + device.channelIndex;
            }
            return name;
        }

        void AddDeviceInDeviceTree(List<TDeviceInfo> listDevice, string mainStationIP)
        {
            deviceDisplayNameMap.Clear();

            TreeNode root = new TreeNode(mainStationIP);
            TreeNode iobox = new TreeNode("I/O Box");
            TreeNode camera = new TreeNode("Camera");
            TreeNode station = new TreeNode("Station");

            iobox.Name = DEVICE_TYPE_NODE;
            camera.Name = DEVICE_TYPE_NODE;
            station.Name = DEVICE_TYPE_NODE;

            TDeviceInfo rootDeviceInfo = new TDeviceInfo();
            rootDeviceInfo.IP = mainStationIP;
            rootDeviceInfo.deviceType = EDeviceType.station;
            rootDeviceInfo.deviceID = "";
            root.Tag = rootDeviceInfo;

            foreach (TDeviceInfo device in listDevice)
            {
                string treeNodename = GetTreeNodeName(device);
                string displayName = device.deviceID + ", " + getReferenceNameSafely(device.refName) + " (" + device.IP + ")";
                if (!deviceDisplayNameMap.ContainsKey(displayName))
                {
                    deviceDisplayNameMap.Add(device.deviceID, displayName);
                }

                TreeNode node = new TreeNode(treeNodename);
                node.Tag = device;
                if (device.deviceType == EDeviceType.camera)
                {
                    node.Name = CAMERA_NODE;
                    camera.Nodes.Add(node);
                    AddDiDoNode(ref node);
                }
                else if (device.deviceType == EDeviceType.station)
                {
                    node.Name = STATION_NODE;
                    station.Nodes.Add(node);
                }
                else if (device.deviceType == EDeviceType.externalDevice)
                {
                    node.Name = IOBOX_NODE;
                    iobox.Nodes.Add(node);
                    AddDiDoNode(ref node);
                }
            }

            root.Nodes.Add(camera);
            root.Nodes.Add(iobox);
            root.Nodes.Add(station);

            treeViewDevices.Nodes.Add(root);
        }

        void ClearDeviceTree()
        {
            treeViewDevices.Nodes.Clear();
        }

        void SetupCmsDeviceTree(List<TDeviceInfo> listDevice)
        {
            List<TDeviceInfo> listStationDevice = new List<TDeviceInfo>();

            foreach (TDeviceInfo device in listDevice)
            {
                if (device.deviceType != EDeviceType.cmsStation) continue;

                listStationDevice.Clear();
                GetDeviceList(listStationDevice, device.deviceID);

                AddDeviceInDeviceTree(listStationDevice, device.refName);
            }
        }

        void SetupDeviceTree(List<TDeviceInfo> listDevice)
        {
            ClearDeviceTree();

            if (serverType == SERVER_TYPE_VAST)
            {
                AddDeviceInDeviceTree(listDevice, textBoxServerIP.Text);
            }
            else if (serverType == SERVER_TYPE_CMS)
            {
                SetupCmsDeviceTree(listDevice);
            }
        }

        unsafe private void GetDeviceList(List<TDeviceInfo> listDevice, string stationID = "")
        {
            TDeviceListReqInfo request = new TDeviceListReqInfo();
            TDeviceListResInfo response = new TDeviceListResInfo();

            request.stationID = stationID;
            request.timeoutMillisecond = 5000;

            int ret = PlatformSDK.GetDeviceList(loginID, ref request, ref response);
            if (ret != 0)
            {
                EventLog.Write("PlatformSDK.GetDeviceList fail: {0:X}", ret);
                return;
            }

            EventLog.Write("PlatformSDK.GetDeviceList success: device num = {0}", response.deviceNum);

            listDevice.Clear();

            for (int i = 0; i < response.deviceNum; ++i)
            {
                IntPtr pDevice = new IntPtr(response.pDeviceList.ToInt64() + Marshal.SizeOf(typeof(TDeviceInfo)) * i);
                TDeviceInfo device = (TDeviceInfo)Marshal.PtrToStructure(pDevice, typeof(TDeviceInfo));
                listDevice.Add(device);
            }
        }

        void GetDeviceListIntoDeviceTree()
        {
            List<TDeviceInfo> listDevice = new List<TDeviceInfo>();
            GetDeviceList(listDevice);
            SetupDeviceTree(listDevice);
        }

        void GetDeviceListWithDetail()
        {
            deviceDetailList.Clear();

            TDeviceListReqInfo request = new TDeviceListReqInfo();
            TDeviceDetailListResponse response = new TDeviceDetailListResponse();

            request.stationID = "";
            request.timeoutMillisecond = 5000;

            int ret = PlatformSDK.GetDeviceListWithDetails(loginID, ref request, ref response);
            if (ret != 0)
            {
                return;
            }

            deviceDetailList.Clear();
            for (int i = 0; i < response.deviceNum; ++i)
            {
                IntPtr pDevice = new IntPtr(response.pDeviceList.ToInt64() + Marshal.SizeOf(typeof(TDeviceDetailInfo)) * i);
                TDeviceDetailInfo device = (TDeviceDetailInfo)Marshal.PtrToStructure(pDevice, typeof(TDeviceDetailInfo));
                deviceDetailList.Add(device);
            }
        }

        private static readonly DateTime dateTime1970 = new DateTime(1970, 1, 1);
        private float[] m_syncSpeed = new float[] { 0.125F, 0.25F, 0.5F, 1F, 2F, 4F, 8F, 16F, 32F, 64F };
        private int currentSyncSpeedIndex = 3;

        uint UtcDateTime2time_t(DateTime datetime)
        {
            TimeSpan diff = datetime.ToUniversalTime() - dateTime1970;
            return (uint)(diff.TotalSeconds);
        }

        uint LocalDateTime2time_t(DateTime datetime)
        {
            TimeSpan diff = datetime.ToLocalTime() - dateTime1970;
            return (uint)(diff.TotalSeconds);
        }

        uint DateTime2time_t(DateTime datetime)
        {
            TimeSpan diff = datetime - dateTime1970;
            return (uint)(diff.TotalSeconds);
        }

        unsafe private void SetLoginInfo(ref TLoginInfo loginInfo)
        {
            loginInfo.IP = textBoxServerIP.Text;
            loginInfo.userName = textBoxServerUser.Text;
            loginInfo.password = textBoxServerPassword.Text;
            loginInfo.port = Convert.ToInt32(textBoxServerPort.Text);
            loginInfo.pfMessCB = messCB;
            loginInfo.context = (IntPtr)gcThis;
            loginInfo.timeoutSeconds = 20;
            loginInfo.blockLogin = false;
            loginInfo.heartBeatLostRetryTimes = MAX_HEARTBEAT_LOST_TIMES;
            loginInfo.heartbeatPeriodSeconds = 5;
            loginInfo.VAST2WebSocketPort = 7768;        //only used for VAST2
            loginInfo.useSSL = checkBoxUseSSL.Checked;
        }

        void GetAlarmListIntoComboAlarm()
        {
            GetAlarmList(alarmList);

            comboBoxAlarm.Items.Clear();

            foreach (AlarmInfo alarm in alarmList)
            {
                string item = alarm.alarmName + " [" + alarm.enabled + "] "
                    + alarm.deviceID + ": " + alarm.DOIndex + " -> ";
                foreach (string cam in alarm.cameraIDList)
                {
                    item += cam;
                    item += " ";
                }
                comboBoxAlarm.Items.Add(item);
            }
        }

        unsafe private void ResetUI()
        {
            loopGetDeviceListTimer.Enabled = false;
            loopUpdateUITimer.Enabled = false;

            configServerLostRetryTimes = 0;

            ClearDeviceTree();
            alarmWindow.Items.Clear();
            eventWindow.Items.Clear();
            listViewBookmark.Clear();
            comboBoxAlarm.Items.Clear();
            alarmList.Clear();
            lock (alarmNotificationLock)
            {
                alarmNotificationList.Clear();
            }
            deviceDisplayNameMap.Clear();
            scheduledChannelTable.Clear();
            recordingDataPanel.ClearUI();
            recordingGroupPanel.ClearUI();

            buttonClosePopupWindow.Text = "Close Popup";

            Interlocked.Exchange(ref bReloginConfigServer, 0);
            Interlocked.Exchange(ref bReloginEventServer, 0);

            didoNodeList.Clear();
            curCheckedDevices.Clear();
        }

        void TestLoopLoginout()
        {
            textBoxServerIP.Text = "127.0.0.1";
            textBoxServerUser.Text = "admin";
            textBoxServerPassword.Text = "admin";
            textBoxServerPort.Text = "3454";

            for (; ; )
            {
                Login();
                while (configServerState != (int)ETunnelState.connected)
                {
                    Thread.Sleep(500);
                }
                Logout();
                UpdateUI();
                Thread.Sleep(1000);
            }
        }

        unsafe private void Login()
        {
            serverType = comboBoxServerType.SelectedIndex;

            TLoginInfo loginInfo = new TLoginInfo();
            SetLoginInfo(ref loginInfo);

            loginID = IntPtr.Zero;
            uint ret = PlatformSDK.Login(ref loginID, ref loginInfo);

            EventLog.Write("PlatformSDK.Login Ret = {0:X}", ret);
            EventLog.Write("PlatformSDK.loginID Ret = {0:X}", loginID);
        }

        unsafe private void Logout()
        {
            ResetUI();

            CloseAllSnapshotChannel();

            uint ret = PlatformSDK.Logout(ref loginID);
            EventLog.Write("PlatformSDK.Logout Ret = {0:X}", ret);
        }

        void UpdateAlarm(string content)
        {
            IntPtr message = Marshal.StringToHGlobalAnsi(content);
            int ret = PlatformSDK.UpdateAlarm(loginID, message);
            EventLog.Write("PlatformSDK.UpdateAlarm Ret = {0:X}", ret);
            Marshal.FreeHGlobal(message);

            GetAlarmListIntoComboAlarm();
        }

        enum EAlarmToggle
        {
            disable,
            enable,
        };

        void SetAlarmToggle(ref string content, EAlarmToggle eToggleState)
        {
            if (eToggleState == EAlarmToggle.enable)
            {
                content = content.Replace("<Enabled>false</Enabled>", "<Enabled>true</Enabled>");
            }
            else
            {
                content = content.Replace("<Enabled>true</Enabled>", "<Enabled>false</Enabled>");
            }
        }

        unsafe private void buttonClickLogin(object sender, EventArgs e)
        {
            Login();
        }

        unsafe private void buttonClickLogout(object sender, EventArgs e)
        {
            Logout();
        }

        void SetDO(TreeNode didoNode)
        {
            TreeNode device = didoNode.Parent;
            if (device == null || device.Tag == null) return;
            if (device.Name != CAMERA_NODE && device.Name != IOBOX_NODE) return;

            int didoID = (int)didoNode.Tag;
            if (!IsDONode(didoID)) return;

            uint index = GetIndexFromDiDoNodeID(didoID);
            uint triggerStatus = GetTriggerStatusFromDiDoNodeID(didoID);
            string deviceID = ((TDeviceInfo)device.Tag).deviceID;
            IntPtr pDeviceID = Marshal.StringToHGlobalAnsi(deviceID);

            triggerStatus = (uint)(triggerStatus == 1 ? 0 : 1);   // invert the trigger status

            const int TIMEOUT_MILSEC = 10000;
            int ret = PlatformSDK.BlockedSetDO(loginID, pDeviceID, index, triggerStatus, TIMEOUT_MILSEC);
            EventLog.Write("PlatformSDK.BlockedSetDO: DO-{1} = {2}, Ret = {0:X}", ret, index, triggerStatus);

            Marshal.FreeHGlobal(pDeviceID);
        }

        void PlayStream(TreeNode camNode)
        {
            string cameraID = ((TDeviceInfo)camNode.Tag).deviceID;

            TCreateChannelInfo info = new TCreateChannelInfo();

            info.serverIP = textBoxServerIP.Text;
            info.serverUserName = textBoxServerUser.Text;
            info.serverPassword = textBoxServerPassword.Text;
            info.serverPort = Convert.ToInt32(checkBoxStreamingOverSSL.Checked ? textBoxStreamingSSLPort.Text : textBoxServerPort.Text);
            info.cameraID = cameraID;
            info.streamIndex = (short)numericUpDownStreamIndex.Value;
            info.capVersion = eCapVersion;
            info.useSSL = checkBoxStreamingOverSSL.Checked;

            ViewCell viewcell = new ViewCell();

            viewcell.CreateLiveChannel(ref info);
            viewcell.Show();
        }

        enum EChannelMode
        {
            live,
            playback,
            export
        };

        void GetLiveSnapshot()
        {
            if (curSelectedDevice.deviceType != EDeviceType.camera) return;
            if (string.IsNullOrEmpty(textBoxDir.Text)) return;
            if (string.IsNullOrEmpty(textBoxFilename.Text)) return;

            TSnapshotContext context = new TSnapshotContext();

            context.savePath = textBoxDir.Text + "\\Live_" +
               System.DateTime.Now.ToString("yyyyMMdd_HHmmss") + ".jpg";

            IntPtr pContext = Marshal.AllocHGlobal(Marshal.SizeOf(context));
            Marshal.StructureToPtr(context, pContext, false);

            TLiveSnapshotRequest request = new TLiveSnapshotRequest();

            request.cameraID = curSelectedDevice.deviceID;
            request.channelIndex = (uint)curSelectedDevice.channelIndex;
            request.streamIndex = (uint)numericUpDownStreamIndex.Value;
            request.pfSnapshotCB = snapshotCB;
            request.width = 0;
            request.height = 0;
            request.IFrameOnly = false;
            request.context = pContext;
            request.timeoutSeconds = 4;

            int ret = PlatformSDK.GetLiveSnapshot(loginID, ref request);
            EventLog.Write("PlatformSDK.GetLiveSnapshot Ret = {0:X}", ret);

            Marshal.FreeHGlobal(pContext);
        }

        void GetPlaybackSnapshot()
        {
            if (curSelectedDevice.deviceType != EDeviceType.camera) return;
            if (string.IsNullOrEmpty(textBoxDir.Text)) return;
            if (string.IsNullOrEmpty(textBoxFilename.Text)) return;

            TSnapshotContext context = new TSnapshotContext();

            context.savePath = textBoxDir.Text + "\\Playback_" +
              dateTimePickerStartTime.Value.ToString("yyyyMMdd_HHmmss") + ".jpg";

            IntPtr pContext = Marshal.AllocHGlobal(Marshal.SizeOf(context));
            Marshal.StructureToPtr(context, pContext, false);

            TPlaybackSnapshotRequest request = new TPlaybackSnapshotRequest();

            request.cameraID = curSelectedDevice.deviceID;
            request.timeSeconds = UtcDateTime2time_t(dateTimePickerStartTime.Value);
            request.millisceondsOfTime = 0;
            request.width = 0;
            request.height = 0;
            request.timeKind = PlatformSDK.ETimeKind.UTC;
            request.timeType = PlatformSDK.ETimeType.recordingTime;
            request.pfSnapshotCB = snapshotCB;
            request.IFrameOnly = false;
            request.context = pContext;
            request.timeoutSeconds = 4;

            int ret = PlatformSDK.GetPlaybackSnapshot(loginID, ref request);
            EventLog.Write("PlatformSDK.GetPlaybackSnapshot Ret = {0:X}", ret);

            Marshal.FreeHGlobal(pContext);
        }

        void LoopGetLiveAndPlaybackChannelSnapshot()
        {
            for (int i = 0; i < 100; ++i)
            {
                foreach (TreeNode device in treeViewDevices.Nodes[0].Nodes[0].Nodes)
                {
                    if (device.Tag == null) continue;
                    curSelectedDevice.deviceID = ((TDeviceInfo)device.Tag).deviceID;
                    GetChannelSnapshot(EChannelMode.live);
                    GetChannelSnapshot(EChannelMode.playback);
                    GetExportFile();
                }
                Thread.Sleep(1000);
            }
        }

        void LoopGetLiveChannelSnapshot()
        {
            for (int i = 0; i < 100; ++i)
            {
                foreach (TreeNode device in treeViewDevices.Nodes[0].Nodes[0].Nodes)
                {
                    if (device.Tag == null) continue;
                    curSelectedDevice.deviceID = ((TDeviceInfo)device.Tag).deviceID;
                    GetChannelSnapshot(EChannelMode.live);
                }
                Thread.Sleep(1000);
            }
        }

        void TestLoopGetLiveChannelSnapshot()
        {
            //            Thread thrd = new Thread(LoopGetLiveAndPlaybackChannelSnapshot);
            //            thrd.Start(); 
        }

        void GetExportFile()
        {
            string cam = curSelectedDevice.deviceID;
            IntPtr exportContext = Marshal.StringToHGlobalAnsi(cam);
            bool useSSL = checkBoxStreamingOverSSL.Checked;

            TExportFileInfo tExportInfo = new TExportFileInfo();

            tExportInfo.startTime = (int)UtcDateTime2time_t(dateTimePickerStartTime.Value);
            tExportInfo.endTime = (int)UtcDateTime2time_t(dateTimePickerEndTime.Value);
            tExportInfo.serverIP = textBoxServerIP.Text;
            tExportInfo.serverUserName = textBoxServerUser.Text;
            tExportInfo.serverPassword = textBoxServerPassword.Text;
            tExportInfo.serverPort = Convert.ToInt32(useSSL ? textBoxStreamingSSLPort.Text : textBoxServerPort.Text);
            tExportInfo.mediaType = PlatformSDK.EMediaType.e3gp;
            tExportInfo.cameraID = cam;
            tExportInfo.substationPath = "";
            tExportInfo.streamIndex = (short)numericUpDownStreamIndex.Value;
            tExportInfo.exportDir = textBoxDir.Text;
            tExportInfo.exportFilename = textBoxFilename.Text;
            tExportInfo.context = exportContext;
            tExportInfo.pfExportStatusCB = exportStatusCB;
            tExportInfo.pfExportInfoCB = exportInfoCB;
            tExportInfo.useSSL = useSSL;

            int timeInterval = tExportInfo.endTime - tExportInfo.startTime;
            if (timeInterval < 0)
            {
                EventLog.Write("Invalid export parameter: timeInterval = {0:X}", timeInterval);
                Marshal.FreeHGlobal(exportContext);
                return;
            }

            ScheduledChannelInfo channelSnapshotInfo = new ScheduledChannelInfo();
            channelSnapshotInfo.context = exportContext;
            channelSnapshotInfo.mode = EChannelMode.export;
            channelSnapshotInfo.waitingSeconds = timeInterval + MAX_SCHEDULED_CHANNEL_WAITING_SECONDS;

            int ret = PlatformSDK.CreateExportMediaFileChannel(ref channelSnapshotInfo.hChannel, ref tExportInfo);
            EventLog.Write("PlatformSDK.CreateExportMediaFileChannel Ret = {0:X}, hChannel = {1:X}"
                , ret, channelSnapshotInfo.hChannel);

            if (ret != 0)
            {
                Marshal.FreeHGlobal(exportContext);
                return;
            }

            newExportingProgress = 0;

            lock (scheduledChannelTableLock)
            {
                scheduledChannelTable.Add(channelSnapshotInfo.hChannel, channelSnapshotInfo);
            }
        }

        void GetChannelSnapshot(EChannelMode mode)
        {
            ScheduledChannelInfo channelSnapshotInfo = new ScheduledChannelInfo();
            channelSnapshotInfo.mode = mode;

            TStreamInfo streamInfo = new TStreamInfo();

            streamInfo.streamSource = PlatformSDK.EStreamSource.VAST;
            streamInfo.streamMode = PlatformSDK.EStreamMode.retrieveOneSnapshot;
            streamInfo.serverIP = textBoxServerIP.Text;
            streamInfo.serverUserName = textBoxServerUser.Text;
            streamInfo.serverPassword = textBoxServerPassword.Text;
            streamInfo.serverPort = Convert.ToInt32(textBoxServerPort.Text);
            streamInfo.cameraID = curSelectedDevice.deviceID;
            streamInfo.decodeType = PlatformSDK.EVideoDecodeType.JPEG;
            streamInfo.pfVideoDecodeCB = videoDecodeCB;
            streamInfo.pfStreamStatusCB = streamStatusCB;
            streamInfo.pfStreamDataCB = null;
            streamInfo.pfPCMDataCB = null;
            streamInfo.pfYV12DataCB = null;
            streamInfo.pfRawDataCB = null;
            streamInfo.hDisplayWindow = IntPtr.Zero;
            streamInfo.streamIndex = (short)numericUpDownStreamIndex.Value;

            int ret = 0;

            if (mode == EChannelMode.live)
            {
                ret = PlatformSDK.CreateLiveChannel(ref channelSnapshotInfo.hChannel, ref streamInfo);
                EventLog.Write(DateTime.Now.ToString("yyyy/MM/dd HH:mm:ss.fff") +
                    "  PlatformSDK.CreateLiveChannel {1:X}, Ret = {0:X}"
                    , ret, channelSnapshotInfo.hChannel);
            }
            else if (mode == EChannelMode.playback)
            {
                streamInfo.startTime = UtcDateTime2time_t(dateTimePickerStartTime.Value);
                streamInfo.endTime = streamInfo.startTime + 3;
                streamInfo.timeType = PlatformSDK.ETimeType.cameraTime;

                ret = PlatformSDK.CreateNetFileChannel(ref channelSnapshotInfo.hChannel, ref streamInfo);
                EventLog.Write(DateTime.Now.ToString("yyyy/MM/dd HH:mm:ss.fff") +
                    "  PlatformSDK.CreateNetFileChannel {1:X}, Ret = {0:X}"
                    , ret, channelSnapshotInfo.hChannel);
            }

            if (ret != 0) return;

            lock (scheduledChannelTableLock)
            {
                scheduledChannelTable.Add(channelSnapshotInfo.hChannel, channelSnapshotInfo);
            }
        }

        bool IsSelectCamera()
        {
            return curSelectedDevice.deviceType == EDeviceType.camera;
        }

        void SetPTZRequest()
        {
            if (!IsSelectCamera()) return;

            ptzRequest.cameraID = curSelectedDevice.deviceID;
            ptzRequest.streamIndex = (uint)numericUpDownStreamIndex.Value;
            ptzRequest.channelIndex = (uint)numericUpDownChannelIndex.Value;
        }

        void SetContinuousPTZRequest()
        {
            if (!IsSelectCamera()) return;

            continuousPTZRequest.cameraID = curSelectedDevice.deviceID;
            continuousPTZRequest.streamIndex = (uint)numericUpDownStreamIndex.Value;
            continuousPTZRequest.channelIndex = (uint)numericUpDownChannelIndex.Value;
        }

        public void PTZCtrl(EPTZCommand cmd, EPTZType type, string modelname)
        {
            if (!IsSelectCamera()) return;

            ptzRequest.PTZType = type;
            ptzRequest.command = cmd;

            int ret = PlatformSDK.PTZControl(loginID, ref ptzRequest);
            EventLog.Write("PlatformSDK.PTZControl Ret = {0:X}", ret);
        }

        public void ContinuousPTZCtrl(int speed, int x, int y)
        {
            if (!IsSelectCamera()) return;

            continuousPTZRequest.speed = speed;
            continuousPTZRequest.x = x;
            continuousPTZRequest.y = y;

            int ret = PlatformSDK.ContinuousPTZControl(loginID, ref continuousPTZRequest);
            EventLog.Write("PlatformSDK.ContinuousPTZControl Ret = {0:X}", ret);
        }

        public void StopPTZCtrl()
        {
            if (!IsSelectCamera()) return;

            int ret = PlatformSDK.StopPTZControl(loginID, ref continuousPTZRequest);
            EventLog.Write("PlatformSDK.StopPTZControl Ret = {0:X}", ret);
        }

        public void PTZMoveToPreset(string presetName)
        {
            if (!IsSelectCamera()) return;

            IntPtr pCamID = Marshal.StringToHGlobalAnsi(ptzRequest.cameraID);
            IntPtr pPresetName = Marshal.StringToHGlobalAnsi(presetName);

            var ret = PlatformSDK.MoveToPreset(loginID, pCamID, pPresetName);
            EventLog.Write("PlatformSDK.MoveToPreset = {0:X}", ret);
        }

        public void GetPTZPresetList()
        {
            if (!IsSelectCamera()) return;

            IntPtr pCamID = Marshal.StringToHGlobalAnsi(ptzRequest.cameraID);
            TSDKPresetList presetList = new TSDKPresetList();

            var ret = PlatformSDK.GetPresetList(loginID, pCamID, ref presetList);
            EventLog.Write("PlatformSDK.GetPresetList = {0:X}, Number of Preset = {1}", ret, presetList.dwPresetNum);

            for (int i = 0; i < presetList.dwPresetNum; ++i)
            {
                IntPtr pPreset = new IntPtr(presetList.ptPresetList.ToInt64() + Marshal.SizeOf(typeof(TSDKPreset)) * i);
                TSDKPreset eventInfo = (TSDKPreset)Marshal.PtrToStructure(pPreset, typeof(TSDKPreset));
                EventLog.Write("Preset Name = {0}", eventInfo.szPresetName);
            }
        }

        private void buttonExport_Click(object sender, EventArgs e)
        {
            if (curSelectedDevice.deviceType != EDeviceType.camera) return;
            GetExportFile();
        }

        private void TestTriggerAllAlarm()
        {
            TManualAlarmInfo alarmInfo = new TManualAlarmInfo();

            foreach (AlarmInfo info in alarmList)
            {
                alarmInfo.ID = info.id;
                int ret = PlatformSDK.TriggerAlarm(loginID, ref alarmInfo);
                EventLog.Write("TriggerAlarm: ID={0}, {1}, {2}, Ret = {3:X}",
                    info.id, info.alarmName, info.deviceID, ret);
            }
        }

        private void buttonTriggerAlarm_Click(object sender, EventArgs e)
        {
            if (comboBoxAlarm.SelectedIndex < 0) return;

            ///{TestTriggerAllAlarm(); return;}

            TManualAlarmInfo alarmInfo = new TManualAlarmInfo();
            alarmInfo.ID = alarmList[comboBoxAlarm.SelectedIndex].id;

            string description = textBoxAlarmDescription.Text;
            if (!string.IsNullOrEmpty(description))
            {
                alarmInfo.szUTF8Description = new byte[PlatformSDK.MAX_ALARM_DESC_LEN + 1];
                byte[] asciiDescription = Encoding.UTF8.GetBytes(description);
                Array.Copy(asciiDescription, alarmInfo.szUTF8Description, asciiDescription.Length);
            }

            int ret = PlatformSDK.TriggerAlarm(loginID, ref alarmInfo);
            EventLog.Write("PlatformSDK.TriggerAlarm Ret = {0:X}", ret);
        }

        private void buttonToggleAlarm_Click(object sender, EventArgs e)
        {
            if (comboBoxAlarm.SelectedIndex < 0) return;

            string content = alarmList[comboBoxAlarm.SelectedIndex].content;

            EAlarmToggle toggleState = (buttonToggleAlarm.Text == "Disable"
                ? EAlarmToggle.disable
                : EAlarmToggle.enable);
            SetAlarmToggle(ref content, toggleState);
            UpdateAlarm(content);
        }

        private void comboBoxAlarm_SelectedIndexChanged(object sender, EventArgs e)
        {
            bool bEnable = alarmList[comboBoxAlarm.SelectedIndex].enabled == "true";
            buttonToggleAlarm.Text = bEnable ? "Disable" : "Enable";
        }

        private void treeViewDevices_NodeMouseDoubeClick(object sender, TreeNodeMouseClickEventArgs e)
        {
            if (e.Node.Tag == null) return;

            if (e.Node.Name == DIDO_NODE)
            {
                SetDO(e.Node);
            }
            else if (e.Node.Name == CAMERA_NODE)
            {
                PlayStream(e.Node);
            }
        }

        private void treeViewDevices_NodeMouseRightClick(object sender, TreeNodeMouseClickEventArgs e)
        {
            if (e.Button != MouseButtons.Right) return;
            if (e.Node.Tag == null) return;

            TDeviceInfo device = (TDeviceInfo)e.Node.Tag;
            if (device.deviceType != EDeviceType.camera && device.deviceType != EDeviceType.station) return;

            foreach (var detail in deviceDetailList)
            {
                if (device.deviceID != detail.deviceID) continue;

                EventLog.Write("================================================");
                EventLog.Write("Model Name        : {0}", detail.modelName);
                EventLog.Write("Firmware Version  : {0}", detail.firmwareVersion);
                EventLog.Write("DI Number         : {0}", detail.DINum);
                EventLog.Write("DO Number         : {0}", detail.DONum);

                if (device.deviceType == EDeviceType.camera)
                {
                    EventLog.Write("MAC               : {0}", detail.mac);
                    EventLog.Write("Stream Number     : {0}", detail.streamNum);
                    EventLog.Write("PTZ Capability    : {0}", detail.ptzCapability);
                    EventLog.Write("Fisheye Capability: {0}", detail.fisheyeCapability);
                    EventLog.Write("Support PTZ: {0}", detail.supportPTZ);
                }
            }
        }

        unsafe private void buttonClosePopupWindow_Click(object sender, EventArgs e)
        {
            if (curSelectedDevice.deviceType != EDeviceType.camera) return;

            int camID = 0;
            string id = curSelectedDevice.deviceID.Substring(2);  //get "xxx" from "C_xxx"
            int.TryParse(id, out camID);

            if (camID == 0) return;

            int ret = PlatformSDK.ClosePopupWindow(loginID, &camID, 1);
            EventLog.Write("PlatformSDK.ClosePopupWindow Ret = {0:X}", ret);
        }

        private void treeViewDevices_AfterSelect(object sender, TreeViewEventArgs e)
        {
            if (e.Node.Tag == null) return;

            if (e.Node.Name == CAMERA_NODE || e.Node.Name == IOBOX_NODE)
            {
                curSelectedDevice = (TDeviceInfo)e.Node.Tag;
            }

            if (e.Node.Name == CAMERA_NODE)
            {
                labelSelectedCameraID.Text = curSelectedDevice.deviceID;

                SetPTZRequest();
                SetContinuousPTZRequest();
            }
        }

        private void treeViewDevices_BeforeCheck(object sender, TreeViewCancelEventArgs e)
        {
            if (e.Node.Tag == null)
            {
                e.Cancel = true;
                return;
            }

            if (e.Node.Name == DIDO_NODE)
            {
                e.Cancel = true;
                return;
            }

            var device = (TDeviceInfo)e.Node.Tag;
            e.Cancel = device.deviceType != EDeviceType.camera;
        }

        private void treeViewDevices_AfterCheck(object sender, TreeViewEventArgs e)
        {
            if (e.Node.Name != CAMERA_NODE) return;

            if (e.Node.Checked)
            {
                curCheckedDevices.Add((TDeviceInfo)e.Node.Tag);
            }
            else
            {
                curCheckedDevices.Remove((TDeviceInfo)e.Node.Tag);
            }
        }

        private void MainUI_Closed(object sender, EventArgs e)
        {
            keepCheckInactiveScheduledChannel = false;
            Logout();
        }

        private void buttonLiveSnapshot_Click(object sender, EventArgs e)
        {
            if (curSelectedDevice.deviceType != EDeviceType.camera) return;
            GetLiveSnapshot();
            //GetChannelSnapshot(EChannelMode.live);
        }

        private void buttonSelectDir_Click(object sender, EventArgs e)
        {
            FolderBrowserDialog dir = new FolderBrowserDialog();
            dir.ShowDialog();
            textBoxDir.Text = dir.SelectedPath;
        }

        private void buttonSelectFilename_Click(object sender, EventArgs e)
        {
            OpenFileDialog file = new OpenFileDialog();
            file.ShowDialog();
            textBoxFilename.Text = file.SafeFileName;
        }

        private void buttonPlayback_Click(object sender, EventArgs e)
        {
            if (curSelectedDevice.deviceType != EDeviceType.camera) return;

            uint startTime = UtcDateTime2time_t(dateTimePickerStartTime.Value);
            uint endTime = UtcDateTime2time_t(dateTimePickerEndTime.Value);
            short streamIndex = (short)numericUpDownStreamIndex.Value;

            TCreateChannelInfo info = new TCreateChannelInfo();

            info.serverIP = textBoxServerIP.Text;
            info.serverUserName = textBoxServerUser.Text;
            info.serverPassword = textBoxServerPassword.Text;
            info.serverPort = Convert.ToInt32(textBoxServerPort.Text);
            info.cameraID = curSelectedDevice.deviceID;
            info.streamIndex = streamIndex;
            info.startTime = startTime;
            info.endTime = endTime;
            info.capVersion = eCapVersion;

            ViewCell viewcell = new ViewCell();
            viewcell.iFrameOnlyInterval = uint.Parse(iFrameTextBox.Text);
            viewcell.CreateNetFileChannel(ref info);
            viewcell.Show();
        }

        private void buttonPlaybackSnapshot_Click(object sender, EventArgs e)
        {
            if (curSelectedDevice.deviceType != EDeviceType.camera) return;
            GetPlaybackSnapshot();
            //GetChannelSnapshot(EChannelMode.playback);            
        }

        private void buttonPTZPanel_Click(object sender, EventArgs e)
        {
            if (ptzPanel.Visible)
            {
                ptzPanel.Hide();
            }
            else
            {
                ptzPanel.Show();
            }
        }

        private void buttonPlayLocalStream_Click(object sender, EventArgs e)
        {
            ViewCell viewcell = new ViewCell();
            viewcell.CreateLocalFileChannel(textBoxDir.Text + "/" + textBoxFilename.Text);
            viewcell.Show();
        }

        private void buttonRecordingList_Click(object sender, EventArgs e)
        {
            if (recordingDataPanel.Visible)
            {
                recordingDataPanel.Hide();
            }
            else
            {
                GetRecordingDeviceList();
                recordingDataPanel.Show();
            }
        }

        private void buttonRecordingGroup_Click(object sender, EventArgs e)
        {
            if (recordingGroupPanel.Visible)
            {
                recordingGroupPanel.Hide();
            }
            else
            {
                recordingGroupPanel.GetRecordingGroupList(loginID);
                recordingGroupPanel.Show();
            }
        }

        private void buttonReloginConfigServer_Click(object sender, EventArgs e)
        {
            Interlocked.Exchange(ref bReloginConfigServer, 1);
            Interlocked.Exchange(ref bDisableGetDeviceList, 1);
            int ret = PlatformSDK.Relogin(loginID, EServerType.configServer);
            EventLog.Write("PlatformSDK.Relogin ConfigServer Ret = {0:X}", ret);
            Interlocked.Exchange(ref bReloginConfigServer, 0);
            if (ret != 0)
            {
                Interlocked.Exchange(ref bReloginConfigServer, 0);
            }
        }

        private void buttonReloginEventServer_Click(object sender, EventArgs e)
        {
            Interlocked.Exchange(ref bReloginEventServer, 1);
            int ret = PlatformSDK.Relogin(loginID, EServerType.eventServer);
            EventLog.Write("PlatformSDK.Relogin EventServer Ret = {0:X}", ret);
            if (ret != 0)
            {
                Interlocked.Exchange(ref bReloginEventServer, 0);
            }
        }

        private void buttonReloginQueryServer_Click(object sender, EventArgs e)
        {
            int ret = PlatformSDK.Relogin(loginID, EServerType.queryServer);
            EventLog.Write("PlatformSDK.Relogin QueryServer Ret = {0:X}", ret);
        }

        private string ComposeGetUserLogHistoryCommandBody()
        {
            return string.Format(
                "<LoginPeriod TZ=\"{0}\"><TmS>{1}</TmS><TmE>{2}</TmE></LoginPeriod>" +
                "<LogoutPeriod TZ=\"{0}\"><TmS>{1}</TmS><TmE>{2}</TmE></LogoutPeriod>"
                , serverTimeZone
                , dateTimePickerStartTime.Value.ToUniversalTime().ToString("yyyy-MM-ddTHH:mm:ssZ")
                , dateTimePickerEndTime.Value.ToUniversalTime().ToString("yyyy-MM-ddTHH:mm:ssZ"));
        }

        void BlockedGetUserLoginHistory()
        {
            IntPtr commandBody = Marshal.StringToHGlobalAnsi(ComposeGetUserLogHistoryCommandBody());

            TBlockedRequestInfo request = new TBlockedRequestInfo();
            TBlockedResponseInfo response = new TBlockedResponseInfo();

            request.commandName = "QryLoginHist";
            request.commandBody = commandBody;
            request.serverType = EServerType.queryServer;
            request.timeoutMillisecond = 3000;

            int ret = PlatformSDK.BlockedSendRequest(loginID, ref request, ref response);
            EventLog.Write("GetUserLoginHistory BlockedSendRequest Ret = {0:X}", ret);

            Marshal.FreeHGlobal(commandBody);

            string str = Marshal.PtrToStringAnsi(response.content);
            EventLog.Write(str);
        }

        void NonBlockedRequestCallBack(IntPtr response, uint responseLen, PlatformSDK.ESDKRequestStatus status, IntPtr context)
        {
            string str = Marshal.PtrToStringAnsi(response);
            EventLog.Write("NonBlockedRequestCallBack status: " + status);
            EventLog.Write(str);
        }

        void NonBlockedGetUserLogHistory()
        {
            IntPtr commandBody = Marshal.StringToHGlobalAnsi(ComposeGetUserLogHistoryCommandBody());

            TNonBlockedRequestInfo request = new TNonBlockedRequestInfo();

            request.commandName = "QryLoginHist";
            request.commandBody = commandBody;
            request.serverType = EServerType.queryServer;
            request.callback = nonBlockedRequestCB;
            request.context = IntPtr.Zero;

            int ret = PlatformSDK.NonBlockedSendRequest(loginID, ref request);
            EventLog.Write("GetUserLogHistory: NonBlockedSendRequest Ret = {0:X}", ret);

            Marshal.FreeHGlobal(commandBody);
        }

        private void buttonUserLogHistory_Click(object sender, EventArgs e)
        {
            BlockedGetUserLoginHistory();
            NonBlockedGetUserLogHistory();
        }

        private void SetVast2BookmarkList(ref TBookmarkList list)
        {
            listViewBookmark.Columns.Clear();

            listViewBookmark.Columns.Add("Bookmark ID", 200, HorizontalAlignment.Left);
            listViewBookmark.Columns.Add("Camera ID", 60, HorizontalAlignment.Left);
            listViewBookmark.Columns.Add("Creator", 60, HorizontalAlignment.Left);
            listViewBookmark.Columns.Add("Describe", 120, HorizontalAlignment.Left);

            listViewBookmark.Items.Clear();

            for (uint i = 0; i < list.bookmarkNum; ++i)
            {
                IntPtr pBookmark = new IntPtr(list.bookmarkList.ToInt64() + Marshal.SizeOf(typeof(TBookmark)) * i);
                TBookmark bookmark = (TBookmark)Marshal.PtrToStructure(pBookmark, typeof(TBookmark));
                ListViewItem lvi = new ListViewItem();

                lvi.Text = bookmark.bookmarkID;
                lvi.SubItems.Add(bookmark.cameraID);
                lvi.SubItems.Add(bookmark.creator);
                lvi.SubItems.Add(bookmark.describe);

                listViewBookmark.Items.Add(lvi);
            }
        }

        private void buttonGetVast2Bookmark_Click(object sender, EventArgs e)
        {
            TBookmarkList list = new TBookmarkList();
            TBookmarkRequest request = new TBookmarkRequest();
            request.dwTimeoutSeconds = 5;

            int ret = PlatformSDK.GetBookmarkList(loginID, ref request, ref list);
            EventLog.Write("PlatformSDK.GetBookmarkList Ret = {0:X}", ret);
            if (ret != 0) return;

            SetVast2BookmarkList(ref list);
            tabControl.SelectedTab = tabPageBookmarkList;
        }

        void HttpRequestCallBack(IntPtr message, uint messageLen, IntPtr context)
        {
            if (messageLen == 0) return;

            string orgmsg = Marshal.PtrToStringAnsi(message);
            string[] lines = System.Text.RegularExpressions.Regex.Split(orgmsg, @"\r?\n|\r");
            string msg = "";
            foreach (string s in lines) msg += (s + Environment.NewLine);

            ResponseMessagePanel resMsgPanel = new ResponseMessagePanel();
            resMsgPanel.SetMessage(msg);
            resMsgPanel.Show();
        }

        private void buttonGetCameraInfo_Click(object sender, EventArgs e)
        {
            string data = "<GetNodeInfo><Node>" + curSelectedDevice.deviceID + "</Node></GetNodeInfo>";
            string url = "http://" + textBoxServerIP.Text + ":" + textBoxServerPort.Text + "/Tunnel/Message.aspx";

            THttpRequest request = new THttpRequest();

            request.method = PlatformSDK.EHttpMethod.POST;
            request.URL = url;
            request.userName = textBoxServerUser.Text;
            request.password = textBoxServerPassword.Text;
            request.header = null;
            request.headerLen = 0;
            request.POSTData = data;
            request.POSTDataLen = (uint)data.Length;
            request.callback = httpRequestCB;
            request.timeoutSeconds = 5;

            int ret = PlatformSDK.SendHttpRequest(ref request);
            EventLog.Write("GetCameraInfo: SendHttpRequest Ret = {0:X}", ret);
        }

        private void checkBoxUseSSL_CheckedChanged(object sender, EventArgs e)
        {
            textBoxServerPort.Text = checkBoxUseSSL.Checked ? "3443" : "3454";
        }

        private void ParseServerTimeZone(string message)
        {
            XmlDocument doc = new XmlDocument();
            doc.LoadXml(message);

            if (doc.GetElementsByTagName("TZ").Count > 0)
            {
                serverTimeZone = doc.GetElementsByTagName("TZ").Item(0).InnerText;
            }
        }

        private void GetServerTimeZone()
        {
            IntPtr commandBody = Marshal.StringToHGlobalAnsi("");

            TBlockedRequestInfo request = new TBlockedRequestInfo();
            TBlockedResponseInfo response = new TBlockedResponseInfo();

            request.commandName = "QryServerTimezone";
            request.commandBody = commandBody;
            request.serverType = EServerType.queryServer;
            request.timeoutMillisecond = 5000;

            int ret = PlatformSDK.BlockedSendRequest(loginID, ref request, ref response);
            EventLog.Write("GetServerTimeZone: BlockedSendRequest Ret = {0:X}", ret);

            Marshal.FreeHGlobal(commandBody);

            if (response.contentLen > 0)
            {
                string message = Marshal.PtrToStringAnsi(response.content);
                ParseServerTimeZone(message);
            }
        }

        private void SetVast1BookmarkList(string message)
        {
            listViewBookmark.Columns.Clear();

            listViewBookmark.Columns.Add("ID", 30, HorizontalAlignment.Left);
            listViewBookmark.Columns.Add("Location", 60, HorizontalAlignment.Left);
            listViewBookmark.Columns.Add("Bookmark Name", 120, HorizontalAlignment.Left);
            listViewBookmark.Columns.Add("Time", 200, HorizontalAlignment.Left);
            listViewBookmark.Columns.Add("Describe", 180, HorizontalAlignment.Left);

            listViewBookmark.Items.Clear();

            //remove namespace, overwise SelectSingleNode should add namespace by XmlNamespaceManager
            string xmlNamespace = " xmlns=\"http://www.vivotek.com/Surveillance/ST7501\"";
            string xmlNoNamespace = message.Replace(xmlNamespace, "");

            XmlDocument doc = new XmlDocument();
            doc.LoadXml(xmlNoNamespace);

            XmlNodeList bookmarkRec = doc.GetElementsByTagName("BookmarkRec");
            if (bookmarkRec.Count == 0) return;

            foreach (XmlNode bookmark in bookmarkRec)
            {
                XmlNode ID = bookmark.SelectSingleNode("ID");
                XmlNode loc = bookmark.SelectSingleNode("Loc");
                XmlNode name = bookmark.SelectSingleNode("Name");
                XmlNode time = bookmark.SelectSingleNode("StaTm");
                XmlNode desc = bookmark.SelectSingleNode("Description");

                ListViewItem lvi = new ListViewItem();

                lvi.Text = ID == null ? "" : ID.InnerText;
                lvi.SubItems.Add(loc == null ? "" : loc.InnerText);
                lvi.SubItems.Add(name == null ? "" : name.InnerText);
                lvi.SubItems.Add(time == null ? "" : time.InnerText);
                lvi.SubItems.Add(desc == null ? "" : desc.InnerText);

                listViewBookmark.Items.Add(lvi);
            }
        }

        private string ComposeGetVast1BookmarkCommandBody()
        {
            return string.Format(
                "<StationGUID>{0}</StationGUID>" +
                "<LocList><Loc>{1}</Loc></LocList>" +
                "<TmBound><TZ>{2}</TZ>" +
                "<StartTm>{3}</StartTm>" +
                "<EndTm>{4}</EndTm></TmBound>" +
                "<TmBase>Station</TmBase><MaxRec>0</MaxRec><Name></Name>"
                , stationID.Substring(2)        //skip the prefix "S_"
                , GetSelectedLocationID()
                , serverTimeZone
                , dateTimePickerStartTime.Value.ToUniversalTime().ToString("yyyy-MM-ddTHH:mm:ssZ")
                , dateTimePickerEndTime.Value.ToUniversalTime().ToString("yyyy-MM-ddTHH:mm:ssZ"));
        }

        private void buttonGetVast1Bookmark_Click(object sender, EventArgs e)
        {
            IntPtr commandBody = Marshal.StringToHGlobalAnsi(ComposeGetVast1BookmarkCommandBody());

            TBlockedRequestInfo request = new TBlockedRequestInfo();
            TBlockedResponseInfo response = new TBlockedResponseInfo();

            request.commandName = "QryBookmarkList";
            request.commandBody = commandBody;
            request.serverType = EServerType.queryServer;
            request.timeoutMillisecond = 10000;

            int ret = PlatformSDK.BlockedSendRequest(loginID, ref request, ref response);
            EventLog.Write("GetVast1Bookmark: BlockedSendRequest Ret = {0:X}", ret);

            Marshal.FreeHGlobal(commandBody);

            string msg = Marshal.PtrToStringAnsi(response.content);
            SetVast1BookmarkList(msg);
        }

        private string GetSelectedLocationID()
        {
            if (comboBoxLocation.SelectedIndex < 0) return "";

            string item = comboBoxLocation.Items[comboBoxLocation.SelectedIndex].ToString();
            return item.Substring(0, item.IndexOf(": "));
        }

        private string ComposeAddVast1BookmarkCommandBody()
        {
            return string.Format(
                "<StationGUID>{0}</StationGUID>" +
                "<Loc>{1}</Loc>" +
                "<Name>{2}</Name>" +
                "<Description>{3}</Description>" +
                "<RecordingTime>{4}</RecordingTime>" +
                "<RecordingTimeMSec>{5}</RecordingTimeMSec>"
                , stationID.Substring(2)        //skip the prefix "S_"
                , GetSelectedLocationID()
                , textBoxBookmarkName.Text
                , textBoxBookmarkDescription.Text
                , UtcDateTime2time_t(dateTimePickerStartTime.Value)
                , 0);
        }

        private void buttonAddVast1Bookmark_Click(object sender, EventArgs e)
        {
            IntPtr commandBody = Marshal.StringToHGlobalAnsi(ComposeAddVast1BookmarkCommandBody());

            TBlockedRequestInfo request = new TBlockedRequestInfo();
            TBlockedResponseInfo response = new TBlockedResponseInfo();

            request.commandName = "AddBookmark";
            request.commandBody = commandBody;
            request.serverType = EServerType.queryServer;
            request.timeoutMillisecond = 10000;

            int ret = PlatformSDK.BlockedSendRequest(loginID, ref request, ref response);
            EventLog.Write("AddVast1Bookmark: BlockedSendRequest Ret = {0:X}", ret);

            Marshal.FreeHGlobal(commandBody);
        }

        private void buttonRemoveBookmark_Click(object sender, EventArgs e)
        {
            if (listViewBookmark.SelectedItems.Count == 0) return;

            string ID = listViewBookmark.SelectedItems[0].SubItems[0].Text;
            IntPtr commandBody = Marshal.StringToHGlobalAnsi("<ID>" + ID + "</ID>");

            TBlockedRequestInfo request = new TBlockedRequestInfo();
            TBlockedResponseInfo response = new TBlockedResponseInfo();

            request.commandName = "DeleteBookmark";
            request.commandBody = commandBody;
            request.serverType = EServerType.queryServer;
            request.timeoutMillisecond = 10000;

            int ret = PlatformSDK.BlockedSendRequest(loginID, ref request, ref response);
            EventLog.Write("RemoveVast1Bookmark: ID = {1}, BlockedSendRequest Ret = {0:X}", ret, ID);

            Marshal.FreeHGlobal(commandBody);
        }

        public void LoadRecordingDeviceList()
        {
            TDeviceListReqInfo request = new TDeviceListReqInfo();
            TRecordingDeviceList deviceList = new TRecordingDeviceList();

            request.stationID = null;
            request.timeoutMillisecond = MAX_GET_REC_DEVICE_WAIT_MILSEC;

            int ret = PlatformSDK.GetRecordingDeviceList(loginID, ref request, ref deviceList);
            EventLog.Write("PlatformSDK.GetRecordingDeviceList Ret = {0:X}", ret);
            if (ret != 0) return;

            comboBoxLocation.Items.Clear();

            for (uint i = 0; i < deviceList.deviceNum; ++i)
            {
                IntPtr pDevice = new IntPtr(deviceList.deviceList.ToInt64() + Marshal.SizeOf(typeof(TRecordingDeviceInfo)) * i);
                TRecordingDeviceInfo device = (TRecordingDeviceInfo)Marshal.PtrToStructure(pDevice, typeof(TRecordingDeviceInfo));

                string item = device.locationID + ": " + getReferenceNameSafely(device.refName);
                comboBoxLocation.Items.Add(item);
            }
        }

        private void tabControl_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (tabControl.SelectedTab == tabPageBookmarkList)
            {
                LoadRecordingDeviceList();
            }
        }

        private void monitorDeviceStatus(object sender, EventArgs e)
        {
            int ret = PlatformSDK.MonitorDeviceStatus(loginID, null, IntPtr.Zero);
            EventLog.Write("PlatformSDK.Stop MonitorDeviceStatus Ret = {0:X}", ret);
        }

        private void stopMonitorEvent(object sender, EventArgs e)
        {
            int ret = PlatformSDK.MonitorEvent(loginID, null, IntPtr.Zero);
            EventLog.Write("PlatformSDK.Stop MonitorEvent Ret = {0:X}", ret);
        }

        private void monitorEvent(object sender, EventArgs e)
        {
            int ret = PlatformSDK.MonitorEvent(loginID, eventCB, IntPtr.Zero);
            EventLog.Write("PlatformSDK.MonitorEvent Ret = {0:X}", ret);
        }

        private void monitorDeviceEvent(object sender, EventArgs e)
        {
            int ret = PlatformSDK.MonitorDeviceStatus(loginID, deviceStatusCB, IntPtr.Zero);
            EventLog.Write("PlatformSDK.MonitorDeviceStatus Ret = {0:X}", ret);
        }

        private void syncPlaybackCB(IntPtr context)
        {
            EventLog.Write("SyncPlaybackCB is called.");
        }

        private void startSyncPlayback(object sender, EventArgs e)
        {
            currentSyncSpeedIndex = 3;
            if (curCheckedDevices.Count <= 0) return;

            uint startTime = UtcDateTime2time_t(dateTimePickerStartTime.Value);
            uint endTime = UtcDateTime2time_t(dateTimePickerEndTime.Value);
            short streamIndex = (short)numericUpDownStreamIndex.Value;

            channelHandles.Clear();
            viewCells.Clear();

            foreach (var node in curCheckedDevices)
            {
                TCreateChannelInfo info = new TCreateChannelInfo();
                info.serverIP = textBoxServerIP.Text;
                info.serverUserName = textBoxServerUser.Text;
                info.serverPassword = textBoxServerPassword.Text;
                info.serverPort = Convert.ToInt32(textBoxServerPort.Text);
                info.cameraID = node.deviceID;
                info.streamIndex = streamIndex;
                info.startTime = startTime;
                info.endTime = endTime;
                info.capVersion = eCapVersion;
                info.useSSL = checkBoxUseSSL.Checked;

                ViewCell viewcell = new ViewCell();
                viewcell.iFrameOnlyInterval = uint.Parse(iFrameTextBox.Text);
                var channel = viewcell.CreateSyncChannel(ref info);

                channelHandles.Add(channel);
                viewCells.Add(viewcell);
            }

            IntPtr hAvSyncPlaybackGroup = IntPtr.Zero;
            int ret11 = PlatformSDK.CreateSyncPlaybackGroup(ref hAvSyncPlaybackGroup);
            EventLog.Write("PlatformSDK.CreateSyncPlaybackGroup Ret = {0:X}", ret11);

            foreach (var handle in channelHandles)
            {
                var ret12 = PlatformSDK.InsertChannelToGroup(hAvSyncPlaybackGroup, handle);
                EventLog.Write("PlatformSDK.InsertChannelToGroup Ret = {0:X}", ret12);
            }

            var option = new PlatformSDK.TSDKSyncPlaybackOption();
            option.startTime = startTime;
            option.startTimeMilSec = 0;
            option.endTime = endTime;
            option.startTimeMilSec = 0;
            option.statusCB = syncPlaybackCB;
            option.statusContext = IntPtr.Zero;
            option.timeKind = PlatformSDK.ETimeKind.UTC;

            var ret3 = PlatformSDK.StartSyncPlaybackGroup(hAvSyncPlaybackGroup, ref option);
            EventLog.Write("PlatformSDK.StartSyncPlaybackGroup Ret = {0:X}", ret3);

            syncPlaybackGroupHandle = hAvSyncPlaybackGroup;

            foreach (var viewCell in viewCells)
            {
                viewCell.syncGroupHandler = hAvSyncPlaybackGroup;
                viewCell.Show();
            }
        }

        private void stopSyncPlayback(object sender, EventArgs e)
        {
            var ret3 = PlatformSDK.StopSyncPlaybackGroup(syncPlaybackGroupHandle);
            EventLog.Write("PlatformSDK.StopSyncPlaybackGroup Ret = {0:X}", ret3);

            foreach (var handle in channelHandles)
            {
                ret3 = PlatformSDK.RemoveChannelFromGroup(syncPlaybackGroupHandle, handle);
                EventLog.Write("PlatformSDK.RemoveChannelFromGroup Ret = {0:X}", ret3);

                ret3 = PlatformSDK.StopChannel(handle);
                EventLog.Write("PlatformSDK.StopChannel Ret = {0:X}", ret3);
            }

            ret3 = PlatformSDK.DeleteSyncPlaybackGroup(ref syncPlaybackGroupHandle);
            EventLog.Write("PlatformSDK.DeleteSyncPlaybackGroup Ret = {0:X}", ret3);
        }

        private void pauseSyncPlayback(object sender, EventArgs e)
        {
            var ret = PlatformSDK.SyncPause(syncPlaybackGroupHandle);
            EventLog.Write("PlatformSDK.SyncPause Ret = {0:X}", ret);
        }

        private void resumeSyncPlayback(object sender, EventArgs e)
        {
            var ret = PlatformSDK.SyncResume(syncPlaybackGroupHandle);
            EventLog.Write("PlatformSDK.SyncResume Ret = {0:X}", ret);
        }

        private void nextFrameSyncPlayback(object sender, EventArgs e)
        {
            var ret = PlatformSDK.SyncNextFrame(syncPlaybackGroupHandle);
            EventLog.Write("PlatformSDK.SyncNextFrame Ret = {0:X}", ret);
        }

        private void speedUpSyncPlayback(object sender, EventArgs e)
        {
            if (m_syncSpeed.Length <= (currentSyncSpeedIndex + 1)) return;
            currentSyncSpeedIndex += 1;

            changeSyncSpeed(m_syncSpeed[currentSyncSpeedIndex]);
        }

        private void speedDownSyncPlayback(object sender, EventArgs e)
        {
            if ((currentSyncSpeedIndex - 1) < 0) return;
            currentSyncSpeedIndex -= 1;

            changeSyncSpeed(m_syncSpeed[currentSyncSpeedIndex]);
        }

        private void handleSeek(object sender, EventArgs e)
        {
            PlatformSDK.TPlaybackOption option = new PlatformSDK.TPlaybackOption();
            option.startTime = UtcDateTime2time_t(dateTimePickerStartTime.Value);
            option.endTime = UtcDateTime2time_t(dateTimePickerEndTime.Value);

            var channelArray = channelHandles.ToArray();
            IntPtr pChannelList = Marshal.AllocHGlobal(channelArray.Length * Marshal.SizeOf(typeof(IntPtr)));
            Marshal.Copy(channelArray, 0, pChannelList, channelArray.Length);

            var channelList = new PlatformSDK.TSDKChannelList();
            channelList.dwChannelNum = (uint)channelHandles.Count;
            channelList.ptChannelList = pChannelList;

            var ret = PlatformSDK.UpdateSyncPlaybackOption(syncPlaybackGroupHandle, ref channelList, ref option);
            EventLog.Write("PlatformSDK.UpdateSyncPlaybackOption Ret = {0:X}", ret);
        }

        private void changeSyncSpeed(float speed)
        {
            uint minimumTime = (uint)1 << 31;
            uint minimumTimeMilisec = 0;
            foreach (var viewCell in viewCells)
            {
                minimumTime = Math.Min(viewCell.frameUTCTimeSecond, minimumTime);

                if (minimumTime == viewCell.frameUTCTimeSecond)
                    minimumTimeMilisec = viewCell.startTimeMilisec;
            }

            var option = new PlatformSDK.TPlaybackOption();
            option.startTime = minimumTime;
            option.startTimeMilisec = minimumTimeMilisec;
            option.endTime = UtcDateTime2time_t(dateTimePickerEndTime.Value);
            option.endTimeMilisec = 0;

            var channelArray = channelHandles.ToArray();
            IntPtr pChannelList = Marshal.AllocHGlobal(channelArray.Length * Marshal.SizeOf(typeof(IntPtr)));
            Marshal.Copy(channelArray, 0, pChannelList, channelArray.Length);

            var channelList = new PlatformSDK.TSDKChannelList();
            channelList.dwChannelNum = (uint)channelHandles.Count;
            channelList.ptChannelList = pChannelList;

            var ret = PlatformSDK.ChangeSyncSpeed(syncPlaybackGroupHandle, ref channelList, speed, ref option);
            EventLog.Write("PlatformSDK.SetSyncSpeed Ret = {0:X} speed = {1}", ret, speed);
        }
        private void PlaybackRawData(object sender, EventArgs e)
        {
            if (curSelectedDevice.deviceType != EDeviceType.camera) return;

            uint startTime = UtcDateTime2time_t(dateTimePickerStartTime.Value);
            uint endTime = UtcDateTime2time_t(dateTimePickerEndTime.Value);
            short streamIndex = (short)numericUpDownStreamIndex.Value;

            TCreateChannelInfo info = new TCreateChannelInfo();

            info.serverIP = textBoxServerIP.Text;
            info.serverUserName = textBoxServerUser.Text;
            info.serverPassword = textBoxServerPassword.Text;
            info.serverPort = Convert.ToInt32(textBoxServerPort.Text);
            info.cameraID = curSelectedDevice.deviceID;
            info.streamIndex = streamIndex;
            info.startTime = startTime;
            info.endTime = endTime;
            info.capVersion = eCapVersion;

            ViewCell viewcell = new ViewCell();
            viewcell.iFrameOnlyInterval = uint.Parse(iFrameTextBox.Text);
            viewcell.CreateNetFileChannel_RawData(ref info);
            viewcell.Show();
        }
    }
}
