﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Runtime.InteropServices;

namespace PlatformSDKSample
{
    using TRecordingDeviceList = PlatformSDK.TRecordingDeviceList;
    using TRecordingDateList = PlatformSDK.TRecordingDateList;
    using TRecordingDateReqInfo = PlatformSDK.TRecordingDateReqInfo;
    using TRecordingDeviceInfo = PlatformSDK.TRecordingDeviceInfo;
    using TRecordingTimeReqInfo = PlatformSDK.TRecordingTimeReqInfo;
    using TDeviceListReqInfo = PlatformSDK.TDeviceListRequest;
    using TRecordingTimeList = PlatformSDK.TRecordingTimeList;
    using TRecordingTime = PlatformSDK.TRecordingTime;
    using TRecordingDate = PlatformSDK.TRecordingDate;
 
    public partial class RecordingDataPanel : Form
    {
        static readonly DateTime dateTime1970 = new DateTime(1970, 1, 1);
        static readonly uint SECONDS_ONE_DAY = 60 * 60 * 24;
        static readonly string CAMERA_NODE = "CAMERA";
        static readonly string DATE_NODE = "DATE";

        private MainUI mainUI;

        const int MAX_GET_REC_DEVICE_WAIT_MILSEC = 10000;
        const int MAX_GET_REC_DATE_WAIT_MILSEC = 5000;

        public RecordingDataPanel(MainUI pMainUI)
        {
            InitializeComponent();

            mainUI = pMainUI;

            InitUI();
        }

        void InitUI()
        {
            this.FormClosing += RecordingPanel_FormClosing;

            timeIntervalWindow.Columns.Add("Index", 40, HorizontalAlignment.Left);
            timeIntervalWindow.Columns.Add("Start Time", 200, HorizontalAlignment.Left);
            timeIntervalWindow.Columns.Add("End Time", 200, HorizontalAlignment.Left);
        }

        public void ClearUI()
        {
            treeViewRecordingDevices.Nodes.Clear();
            timeIntervalWindow.Items.Clear();
        }

        void RecordingPanel_FormClosing(object sender, FormClosingEventArgs e)
        {
            this.Hide();
            e.Cancel = true; // this cancels the close event.
        }

        uint UtcDateTime2time_t(DateTime datetime)
        {
            TimeSpan diff = datetime.ToUniversalTime() - dateTime1970;
            return (uint)(diff.TotalSeconds);
        }

        private string getReferenceNameSafely(in string refName)
        {
            return refName.Length < 2 ? "" : refName.Substring(2);
        }

        public void SetRecordingDateList(TreeNode camera, ref TRecordingDateList dateList)
        {
            camera.Nodes.Clear();

            for (uint j = 0; j < dateList.dateNum; ++j)
            {
                IntPtr pDate = new IntPtr(dateList.dateList.ToInt64() + Marshal.SizeOf(typeof(TRecordingDate)) * j);
                TRecordingDate recDate = (TRecordingDate)Marshal.PtrToStructure(pDate, typeof(TRecordingDate));

                TreeNode date = new TreeNode(recDate.date);
                date.Tag = camera.Tag;
                date.Name = DATE_NODE;
                camera.Nodes.Add(date);
            }
        }

        public void SetRecordingDeviceList(ref TRecordingDeviceList deviceList)
        {
            TreeNode root = new TreeNode("Host");

            for (uint i = 0; i < deviceList.deviceNum; ++i)
            {
                IntPtr pDevice = new IntPtr(deviceList.deviceList.ToInt64() + Marshal.SizeOf(typeof(TRecordingDeviceInfo)) * i);
                TRecordingDeviceInfo device = (TRecordingDeviceInfo)Marshal.PtrToStructure(pDevice, typeof(TRecordingDeviceInfo));               

                string treeNodename = device.deviceID + ", " + getReferenceNameSafely(device.refName);
                TreeNode camera = new TreeNode(treeNodename);
                camera.Tag = device.locationID;
                camera.Name = CAMERA_NODE;
                
                root.Nodes.Add(camera);
            }

            treeViewRecordingDevices.Nodes.Clear();
            treeViewRecordingDevices.Nodes.Add(root);
        }

        void AddTimeIntervalInList(ref TRecordingTime timeInfo)
        {
            long startLocalSeconds = timeInfo.tServerTime.startSecond 
                + timeInfo.tServerTime.startTimeZoneSecond 
                + timeInfo.tServerTime.startDayLightSavingSecond;

            long endLocalSeconds = timeInfo.tServerTime.endSecond
                + timeInfo.tServerTime.endTimeZoneSecond
                + timeInfo.tServerTime.endDayLightSavingSecond;

            DateTime startDate = new DateTime(1970, 1, 1, 0, 0, 0, 0).AddSeconds(startLocalSeconds);
            DateTime endDate = new DateTime(1970, 1, 1, 0, 0, 0, 0).AddSeconds(endLocalSeconds);     
      
            ListViewItem lvi = new ListViewItem();
            
            lvi.Text = (timeIntervalWindow.Items.Count + 1).ToString();
            lvi.SubItems.Add(startDate.ToString("yyyy-MM-dd HH:mm:ss"));
            lvi.SubItems.Add(endDate.ToString("yyyy-MM-dd HH:mm:ss"));

            timeIntervalWindow.Items.Add(lvi);
        }

        void SetRecordingTimeList(ref TRecordingTimeList recTimeList)
        {
            timeIntervalWindow.Items.Clear();

            for (uint i = 0; i < recTimeList.timeNum; ++i)
            {
                IntPtr pTime = new IntPtr(recTimeList.timeList.ToInt64() + Marshal.SizeOf(typeof(TRecordingTime)) * i);
                TRecordingTime time = (TRecordingTime)Marshal.PtrToStructure(pTime, typeof(TRecordingTime));      

                AddTimeIntervalInList(ref time);
            }
        }

        void GetRecordingTimeList(string locationID, uint utcStartTimeSeconds, uint utcEndTimeSeconds)
        {
            TRecordingTimeReqInfo request = new TRecordingTimeReqInfo();
            TRecordingTimeList timeList = new TRecordingTimeList();

            request.utcStartTimeSeconds = utcStartTimeSeconds;
            request.utcEndTimeSeconds = utcEndTimeSeconds;
            request.locationID = locationID;
            request.timeoutMillisecond = 5000;

            IntPtr pLocationID = Marshal.StringToHGlobalAnsi(locationID);

            var ret = PlatformSDK.GetRecordingTimeList(mainUI.loginID, ref request, ref timeList);
            Console.WriteLine("PlatformSDK.GetRecordingTimeList Ret = {0:X}", ret);

            Marshal.FreeHGlobal(pLocationID);

            SetRecordingTimeList(ref timeList);
        }

        void GetRecordingDateList(TreeNode camera)
        {
            TRecordingDateReqInfo request = new TRecordingDateReqInfo();
            TRecordingDateList dateList = new TRecordingDateList();

            uint startTime = UtcDateTime2time_t(recDataStartTime.Value);
            uint endTime = UtcDateTime2time_t(recDataEndTime.Value);
            request.locationID = (string) camera.Tag;
            request.timeoutMillisecond = MAX_GET_REC_DATE_WAIT_MILSEC;
            request.utcStartTimeSeconds = startTime;
            request.utcEndTimeSeconds = endTime;

            IntPtr pLocationID = Marshal.StringToHGlobalAnsi(request.locationID);

            uint ret = PlatformSDK.GetRecordingDateList(mainUI.loginID, ref request, ref dateList);
            Console.WriteLine("PlatformSDK.GetRecordingDateList Ret = {0:X}", ret);

            Marshal.FreeHGlobal(pLocationID);

            SetRecordingDateList(camera, ref dateList);
        } 

        void treeViewRecordingDevices_NodeMouseDoubeClick(object sender, TreeNodeMouseClickEventArgs e)
        {
            if (e.Node.Tag == null) return;

            if (e.Node.Name == CAMERA_NODE)
            {
                GetRecordingDateList(e.Node);
            }
            else if (e.Node.Name == DATE_NODE)
            {
                int[] part = StringUtility.ToIntArray(e.Node.Text, '-');
                DateTime date = new DateTime(part[0], part[1], part[2], 0, 0, 0, 0);
                TimeSpan diff = date - dateTime1970;
                uint startUtcSeconds = (uint)diff.TotalSeconds;
                GetRecordingTimeList((string)e.Node.Tag, startUtcSeconds, startUtcSeconds + SECONDS_ONE_DAY);
            }
        }
    }

    public static class StringUtility
    {
        public static int[] ToIntArray(this string value, char separator)
        {
            return Array.ConvertAll(value.Split(separator), s => int.Parse(s));
        }
    }
}
