﻿using System;
using System.Collections.Generic;
using System.Collections;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Runtime.InteropServices;

namespace PlatformSDKSample
{
    using EManualRecordingAction = PlatformSDK.EManualRecordingAction;
    using TManualRecordingRequestDevice = PlatformSDK.TManualRecordingRequestDevice;
    using TManualRecordingRequest = PlatformSDK.TManualRecordingRequest;
    using TRecordingGroupDeviceInfo = PlatformSDK.TRecordingGroupDeviceInfo;
    using TRecordingGroupPathInfo = PlatformSDK.TRecordingGroupPathInfo;
    using TRecordingGroupReqInfo = PlatformSDK.TRecordingGroupReqInfo;
    using TRecordingGroupList = PlatformSDK.TRecordingGroupList;
    using TRecordingGroup = PlatformSDK.TRecordingGroup;

    public partial class RecordingGroupPanel : Form
    {
        static readonly string GROUP_NODE = "GROUP";
        static readonly string PATH_NODE = "PATH";
        static readonly string CAMERA_NODE = "CAMERA";

        const int MAX_GET_REC_GROUP_WAIT_MILSEC = 10000;
        const int MAX_SET_MANUAL_REC_WAIT_MILSEC = 5000;
        const int ATTR_WINDOW_NAME_FIELD_WIDTH = 130;

        List<TRecordingGroupDeviceInfo> deviceInfoList;
        List<TRecordingGroupPathInfo> pathInfoList;
        List<TRecordingGroup> groupInfoList;
        MainUI mainUI;
        IntPtr loginID;
        uint SelectedGroupID = 0;
        
        public RecordingGroupPanel(MainUI pMainUI)
        {
            InitializeComponent();

            mainUI = pMainUI;
            deviceInfoList = new List<TRecordingGroupDeviceInfo>();
            pathInfoList = new List<TRecordingGroupPathInfo>();
            groupInfoList = new List<TRecordingGroup>();

            InitUI();
        }

        void InitUI()
        {
            this.FormClosing += RecordingGroupPanel_FormClosing;

            int valueWidth = recordingGroupAttributeWindow.Width - ATTR_WINDOW_NAME_FIELD_WIDTH - 4;
            recordingGroupAttributeWindow.Columns.Add("Name", ATTR_WINDOW_NAME_FIELD_WIDTH, HorizontalAlignment.Left);
            recordingGroupAttributeWindow.Columns.Add("Value", valueWidth, HorizontalAlignment.Left);
        }

        void RecordingGroupPanel_FormClosing(object sender, FormClosingEventArgs e)
        {
            this.Hide();
            e.Cancel = true; // this cancels the close event.
        }

        public void ClearUI()
        {
            treeViewRecordingGroupDevices.Nodes.Clear();
            recordingGroupAttributeWindow.Items.Clear();
        }

        void RecordingPanel_FormClosing(object sender, FormClosingEventArgs e)
        {
            this.Hide();
            e.Cancel = true; // this cancels the close event.
        }
        private string getReferenceNameSafely(in string refName)
        {
            return refName.Length < 2 ? "" : refName.Substring(2);
        }

        public void GetRecordingGroupList(IntPtr mainUILoginID)
        {
            loginID = mainUILoginID;

            TRecordingGroupReqInfo request = new TRecordingGroupReqInfo();
            TRecordingGroupList groupList = new TRecordingGroupList();

            request.stationID = null;
            request.timeoutMillisecond = MAX_GET_REC_GROUP_WAIT_MILSEC;

            int ret = PlatformSDK.GetRecordingGroup(loginID, ref request, ref groupList);
            Console.WriteLine("PlatformSDK.GetRecordingGroup Ret = {0:X}", ret);

            SetRecordingGroupList(ref groupList);
        }

        public void SetRecordingGroupList(ref TRecordingGroupList groupList)
        {
            treeViewRecordingGroupDevices.Nodes.Clear();
            deviceInfoList.Clear();
            groupInfoList.Clear();

            TreeNode root = new TreeNode("Host");

            for (uint i = 0; i < groupList.groupNum; ++i)
            {
                IntPtr pGroup = new IntPtr(groupList.groupList.ToInt64() + Marshal.SizeOf(typeof(TRecordingGroup)) * i);
                TRecordingGroup group = (TRecordingGroup)Marshal.PtrToStructure(pGroup, typeof(TRecordingGroup));

                string treeNodename = group.groupName;
                TreeNode groupNode = new TreeNode(treeNodename);
                groupNode.Tag = groupInfoList.Count;
                groupNode.Name = GROUP_NODE;

                root.Nodes.Add(groupNode);
                groupInfoList.Add(group);

                TreeNode pathNode = new TreeNode("Path");
                TreeNode deviceNode = new TreeNode("Camera");

                groupNode.Nodes.Add(pathNode);
                groupNode.Nodes.Add(deviceNode);

                for (uint j = 0; j < group.pathNum; ++j)
                {
                    IntPtr pPath = new IntPtr(group.pathList.ToInt64() + Marshal.SizeOf(typeof(TRecordingGroupPathInfo)) * j);
                    TRecordingGroupPathInfo path = (TRecordingGroupPathInfo)Marshal.PtrToStructure(pPath, typeof(TRecordingGroupPathInfo));
                    
                    TreeNode recPath = new TreeNode(path.path + " (" + path.reserveSpace + " MB)");
                    recPath.Tag = pathInfoList.Count;
                    recPath.Name = PATH_NODE;

                    pathInfoList.Add(path);
                    pathNode.Nodes.Add(recPath);
                }

                for (uint j = 0; j < group.deviceNum; ++j)
                {
                    IntPtr pDevice = new IntPtr(group.deviceList.ToInt64() + Marshal.SizeOf(typeof(TRecordingGroupDeviceInfo)) * j);
                    TRecordingGroupDeviceInfo device = (TRecordingGroupDeviceInfo)Marshal.PtrToStructure(pDevice, typeof(TRecordingGroupDeviceInfo));

                    TreeNode camera = new TreeNode(getReferenceNameSafely(device.refName) + " (" + device.deviceID + ")");
                    camera.Tag = deviceInfoList.Count;
                    camera.Name = CAMERA_NODE;

                    deviceInfoList.Add(device);
                    deviceNode.Nodes.Add(camera);
                }
            }

            treeViewRecordingGroupDevices.Nodes.Clear();
            treeViewRecordingGroupDevices.Nodes.Add(root);
        }

        void treeViewRecordingGroupDevices_NodeMouseClick(object sender, TreeNodeMouseClickEventArgs e)
        {
            if (e.Node.Tag == null) return;

            if (e.Node.Name == CAMERA_NODE)
            {
                SelectedGroupID = groupInfoList[(int)e.Node.Parent.Parent.Tag].groupID;
                int deviceInfoListIndex = (int)e.Node.Tag;
                labelRecGroupSelectedDevice.Text = deviceInfoList[deviceInfoListIndex].deviceID;
                SetDeviceAttribute(deviceInfoListIndex);
            }
            else if (e.Node.Name == PATH_NODE)
            {
                SetPathAttribute((int)e.Node.Tag);
            }
            else if (e.Node.Name == GROUP_NODE)
            {
                SetGroupAttribute((int)e.Node.Tag);
            }
        }

        void AddAttributeItem(string name, string value)
        {
            ListViewItem item = new ListViewItem();
            item.Text = name;
            item.SubItems.Add(value);
            recordingGroupAttributeWindow.Items.Add(item);
        }

        void SetDeviceAttribute(int deviceInfoListIndex)
        {
            recordingGroupAttributeWindow.Items.Clear();

            TRecordingGroupDeviceInfo device = deviceInfoList[deviceInfoListIndex];
            AddAttributeItem("Camera ID", device.deviceID);
            AddAttributeItem("Camera Reference Name", getReferenceNameSafely(device.refName));
            AddAttributeItem("Location ID", device.locationID);
            AddAttributeItem("Stream Index", device.streamIndex.ToString());
            AddAttributeItem("Pre-event Seconds", device.preEventSecond.ToString());
            AddAttributeItem("Post-event Seconds", device.postEventSecond.ToString());
            AddAttributeItem("Enable Key Frame Only", device.enableKeyFrameOnly.ToString());
            AddAttributeItem("Enable Seamless", device.enableSeamless.ToString());
            AddAttributeItem("Enable TimeShift", device.enableTimeShift.ToString());
        }

        void SetPathAttribute(int pathInfoListIndex)
        {
            recordingGroupAttributeWindow.Items.Clear();

            TRecordingGroupPathInfo path = pathInfoList[pathInfoListIndex];

            string type = "Unknown";
            if (path.type == PlatformSDK.ERecordingPathType.local)
            {
                type = "Local";
            }
            else if (path.type == PlatformSDK.ERecordingPathType.networkDomain)
            {
                type = "Network Domain";
            }
            else if (path.type == PlatformSDK.ERecordingPathType.networkHost)
            {
                type = "Network Host";
            }

            AddAttributeItem("Path", path.path);
            AddAttributeItem("Type", type);
            AddAttributeItem("Reserve Space", path.reserveSpace.ToString() + " MB");
            AddAttributeItem("Status", path.bConnected ? "Connected" : "Disconnected");

            if (!string.IsNullOrEmpty(path.hostDomain))
            {
                AddAttributeItem("Host Domain", path.hostDomain);
            }
            if (!string.IsNullOrEmpty(path.hostUsername))
            {
                AddAttributeItem("Host Username", path.hostUsername);
            }
            if (!string.IsNullOrEmpty(path.hostPassword))
            {
                AddAttributeItem("Host Password", path.hostPassword);
            }
        }

        void SetGroupAttribute(int groupInfoListIndex)
        {
            recordingGroupAttributeWindow.Items.Clear();

            TRecordingGroup group = groupInfoList[groupInfoListIndex];
            AddAttributeItem("Name", group.groupName);
            AddAttributeItem("ID", group.groupID.ToString());
            AddAttributeItem("Enable Time-based Cycle", group.enableTimeBaseCycle.ToString());
            AddAttributeItem("Reserve Cycle Time", group.reserveCycleMinute.ToString() + " Minutes");
        }

        void SetManualRecording(EManualRecordingAction eAction)
        {
            TManualRecordingRequest request = new TManualRecordingRequest();
            TManualRecordingRequestDevice device = new TManualRecordingRequestDevice();

            device.deviceID = labelRecGroupSelectedDevice.Text;
            device.groupID = SelectedGroupID;

            IntPtr pDevice = Marshal.AllocHGlobal(Marshal.SizeOf(device));
            Marshal.StructureToPtr(device, pDevice, false);

            request.eRecAction = eAction;
            request.deviceNum = 1;
            request.deviceList = pDevice;
            request.timeoutMillisecond = MAX_SET_MANUAL_REC_WAIT_MILSEC;

            try
            {
                int ret = PlatformSDK.SetManualRecording(loginID, ref request);
                Console.WriteLine("PlatformSDK.GetRecordingGroup Ret = {0:X}", ret);
            }
            finally
            {
                Marshal.FreeHGlobal(pDevice);
            }
        }

        void buttonManualRecording_Click(object sender, EventArgs e)
        {
            SetManualRecording(EManualRecordingAction.start);
        }

        private void buttonStopManualRecoding_Click(object sender, EventArgs e)
        {
            SetManualRecording(EManualRecordingAction.stop);
        }
    }
}
