﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Threading;
using System.Runtime.InteropServices;
using System.IO;
using System.Drawing.Imaging;

namespace PlatformSDKSample
{
    using TVideoDecodeInfo = PlatformSDK.TVideoDecodeInfo;
    using TStreamInfo = PlatformSDK.TStreamInfo;
    using TStreamStatusInfo = PlatformSDK.TStreamStatusInfo;
    using TLocalFileInfo = PlatformSDK.TLocalFileInfo;
    using TStreamDataInfo = PlatformSDK.TStreamDataInfo;
    using TStreamingDisplayInfo = PlatformSDK.TStreamingDisplayInfo;
    using EPlayDirection = PlatformSDK.EPlayDirection;
    using ECapabilityVersion = PlatformSDK.ECapabilityVersion;
    using EHardwareAcceleration = PlatformSDK.EHardwareAcceleration;
    
    public partial class ViewCell : Form
    {
        enum EChannelType
        {
            unknown,
            live,
            playback,
            localFile,
        }

        enum EStreamState
        {
            stop,
            play,
            pause,
        }

        const string forwardPlayText = "►";
        const string backwardPlayText = "◄";
        const string pauseText = "∥";
        const string stopText = "▣";

        const float MAX_SPEED = 64;
        const float MIN_SPEED = 1.0f / 8;

        uint startTime = 0;
        uint endTime = 0;

        EPlayDirection playDirection = EPlayDirection.forward;
        PlatformSDK.pfStreamDataCallBack streamDataCB = null;
        PlatformSDK.pfVideoDecodeCallBack videoDecodeCB = null;
        PlatformSDK.pfStreamStatusCallBack streamStatusCB = null;
        PlatformSDK.pfYV12DataCallBack yv12DataCallBack = null;
        PlatformSDK.pfRawDataCallBack rawDataCallBack = null;

        FrameCounter frameCounter = new FrameCounter();

        TCreateChannelInfo reCreateChannelInfo = new TCreateChannelInfo();
        EChannelType channelType = EChannelType.unknown;
        EStreamState streamState = EStreamState.stop;
        public IntPtr syncGroupHandler = IntPtr.Zero;
        IntPtr hChannel = IntPtr.Zero;
        float speed = 1;
        bool turn = true;
        public uint frameUTCTimeSecond = 0;
        public uint iFrameOnlyInterval = 0;
        public uint startTimeMilisec = 0;
        private bool isSyncPlayback = false;

        struct TImageInfo
        {
            public byte[] image;
            public uint frameSecond;
            public uint frameMilSec; 
        }
        readonly object imageQueueLock = new Object();
        readonly object imageLock = new Object();
        System.Timers.Timer timerLoopShowSnapshot = new System.Timers.Timer(20);
        Queue<TImageInfo> imageQueue = new Queue<TImageInfo>();

        Font drawFont = new Font("Arial", 10);
        SolidBrush drawBrush = new SolidBrush(Color.Red);

        void ShowSnapshot(object source, System.Timers.ElapsedEventArgs e)
        {
            TImageInfo img;
            lock (imageQueueLock)
            {
                if (imageQueue.Count == 0) return;
                img = imageQueue.Dequeue();
            }
            Image image = Image.FromStream(new MemoryStream(img.image));
            Graphics g = Graphics.FromImage(image);
            
            System.DateTime time = new System.DateTime(1970, 1, 1).AddSeconds(img.frameSecond);
            String timeString = time.ToString() + "." + img.frameMilSec + "  FPS = " + frameCounter.getFPS();
                      
            panelViewCell.BackgroundImage = image;
        }        

        void TestLoopShowSnapshot()
        {
            timerLoopShowSnapshot.Elapsed += new System.Timers.ElapsedEventHandler(ShowSnapshot);
            timerLoopShowSnapshot.AutoReset = true;
            timerLoopShowSnapshot.Enabled = true;
        }

        public ViewCell()
        {
            InitializeComponent();
            SetStyle(ControlStyles.SupportsTransparentBackColor, true);
            streamDataCB = new PlatformSDK.pfStreamDataCallBack(StreamDataCallBack);
            videoDecodeCB = new PlatformSDK.pfVideoDecodeCallBack(VideoDecodeCallBack);
            streamStatusCB = new PlatformSDK.pfStreamStatusCallBack(StreamStatusCallBack);
            yv12DataCallBack = new PlatformSDK.pfYV12DataCallBack(YV12DataCallBack);
            rawDataCallBack = new PlatformSDK.pfRawDataCallBack(RawDataCallBack);
        }

        ~ViewCell()
        {
            DeleteChannel();
            imageQueue.Clear();
            GC.Collect();            
            GC.WaitForPendingFinalizers();
        }

        void VideoDecodeCallBack(IntPtr pVideoDecodeInfo)
        {
            TVideoDecodeInfo decodeInfo = (TVideoDecodeInfo)Marshal.PtrToStructure(pVideoDecodeInfo, typeof(TVideoDecodeInfo));

            if (decodeInfo.pBuffer == IntPtr.Zero) return;

            TImageInfo image = new TImageInfo();
            image.frameSecond = decodeInfo.frameSecond;
            image.frameMilSec = decodeInfo.frameMilSec;
            image.image = new byte[decodeInfo.bufSize];
            Marshal.Copy(decodeInfo.pBuffer, image.image, 0, decodeInfo.bufSize);

            lock (imageQueueLock)
            {
                imageQueue.Enqueue(image);
            }

            Marshal.FreeHGlobal(decodeInfo.pBuffer);
        }
        void YV12DataCallBack(IntPtr hChannel, IntPtr pBuffer, uint bufferSize, int height, int width, int dataType, IntPtr context)
        {
            if (pBuffer == IntPtr.Zero)
            {
                return;
            }

            byte[] frameBuffer = new byte[bufferSize];
            Marshal.Copy(pBuffer, frameBuffer, 0, frameBuffer.Length);

            EventLog.Write("YV12DataCallBack Ret = {0} {1} {2} {3} {4}", hChannel, bufferSize, height, width, dataType);
        }

        Bitmap GetDataPicture(int w, int h, byte[] data)
        {
            Bitmap pic = new Bitmap(w, h, System.Drawing.Imaging.PixelFormat.Format32bppRgb);

            for (int y = 0; y < h; y++)
            {
                for (int x = 0; x < w; x++)
                {
                    int arrayIndex = (y * w + x) * 3;
                    Color c = Color.FromArgb(
                       data[arrayIndex + 2],
                       data[arrayIndex + 1],
                       data[arrayIndex]
                    );
                    pic.SetPixel(x, y, c);
                }
            }

            return pic;
        }

        public Bitmap ResizeBitmap(Bitmap bmp, int width, int height)
        {
            Bitmap result = new Bitmap(width, height);
            using (Graphics g = Graphics.FromImage(result))
            {
                g.DrawImage(bmp, 0, 0, width, height);
            }

            return result;
        }

        void ShowRawDataOnViewcell(IntPtr pBuffer, uint bufferSize, int height, int width)
        {
            byte[] image = new byte[bufferSize];
            Marshal.Copy(pBuffer, image, 0, (int)bufferSize);

            Bitmap pic = GetDataPicture(width, height, image);
            Bitmap newImage = ResizeBitmap(pic, panelViewCell.Width, panelViewCell.Height);
            panelViewCell.BackgroundImage = newImage;
        }

        void RawDataCallBack(IntPtr hChannel, IntPtr pBuffer, uint bufferSize, int height, int width, uint frameSecond, uint frameMSec)
        {
            if (pBuffer == IntPtr.Zero)
            {
                return;
            }
            EventLog.Write("RawDataCallBack Ret = {0} {1} {2} {3} time:{4}.{5}", hChannel, bufferSize, height, width, frameSecond, frameMSec);

            ShowRawDataOnViewcell(pBuffer, bufferSize, height, width);
        }

        void StreamDataCallBack(IntPtr pStreamDataInfo)
        {            
            TStreamDataInfo info = (TStreamDataInfo)Marshal.PtrToStructure(pStreamDataInfo, typeof(TStreamDataInfo));
            frameUTCTimeSecond = info.frameUTCTimeSecond;
            startTimeMilisec = info.frameUTCTimeMilSec;

            EventLog.Write("StreamDataCallBack codecType: {0} {1} {2}", info.codecType, frameUTCTimeSecond, startTimeMilisec);

            try
            {
                this.BeginInvoke((MethodInvoker)delegate ()
                {
                    var isDummyFrame = info.dummyPacket;
                    dummyPanel.Visible = isDummyFrame;
                });
            }
            catch (System.ObjectDisposedException)
            {
                EventLog.Write("ViewCell is disposed");
            }
            catch (System.InvalidOperationException)
            {
                EventLog.Write("ViewCell is not created");
            }
            //System.DateTime time = new System.DateTime(1970, 1, 1).AddSeconds(info.frameLocalTimeSecond);
            //Console.WriteLine(time.ToString() + "." + info.frameLocalTimeMilSec);
        }

        void StreamStatusCallBack(IntPtr pStatusInfo)
        {
            TStreamStatusInfo info = (TStreamStatusInfo)Marshal.PtrToStructure(pStatusInfo, typeof(TStreamStatusInfo));
            EventLog.Write("StreamStatusCallBack: hChannel = {0:X}, Status = {1}", info.hChannel, info.status);

            if (isSyncPlayback && info.status == PlatformSDK.EStreamStatus.stopped)
            {
                PlatformSDK.TPlaybackOption option = new PlatformSDK.TPlaybackOption();
                option.startTime = frameUTCTimeSecond;
                option.startTimeMilisec = startTimeMilisec;
                option.endTime = endTime;

                var ret = PlatformSDK.ReconnectSyncPlaybackChannel(syncGroupHandler, info.hChannel, ref option);
                EventLog.Write("PlatformSDK.ReconnectSyncPlaybackChannel Ret = {0:X}", ret);
                info.status = (ret == 0) ? PlatformSDK.EStreamStatus.ok : PlatformSDK.EStreamStatus.stopped;
            }

            try
            {
                this.Invoke((MethodInvoker)delegate ()
                {
                    if (info.status == PlatformSDK.EStreamStatus.stopped)
                    {
                        DisconnectPanel.Visible = true;
                        ReconnectButton.Show();
                    }
                    else
                    {
                        DisconnectPanel.Visible = false;
                        ReconnectButton.Hide();
                    }
                });
            }
            catch (System.ObjectDisposedException)
            {
                EventLog.Write("ViewCell is disposed");
            }
            catch (System.InvalidOperationException)
            {
                EventLog.Write("ViewCell is not created");
            }
        }

        void SetDisplayInfo()
        {
            TStreamingDisplayInfo info = new TStreamingDisplayInfo();

            info.showFrameTime = true;
            info.showMillisecond = true;
            info.showTextShadow = true;
            info.opaqueBackground = true; 
            info.timeSource = PlatformSDK.ETimeSource.recordingTime;
            info.timeKind = PlatformSDK.ETimeKind.local;
            info.ARGBTextColor = (0<<16) | (0<<8) | 255;
            info.ARGBBackgroundColor = (0 << 16) | (255 << 8) | 0;
            info.fontSize = 22;

            int ret = PlatformSDK.SetDisplayInfo(hChannel, ref info);
            EventLog.Write("SetDisplayInfo: {0:X}", ret);
        }

        void CreateLiveChannel(TCreateChannelInfo info)
        {
            reCreateChannelInfo = info;
            TStreamInfo streamInfo = new TStreamInfo();

            streamInfo.streamSource = PlatformSDK.EStreamSource.VAST;
            streamInfo.streamMode = PlatformSDK.EStreamMode.normalStream;
            streamInfo.serverIP = info.serverIP;
            streamInfo.serverUserName = info.serverUserName;
            streamInfo.serverPassword = info.serverPassword;
            streamInfo.serverPort = Convert.ToInt32(info.serverPort);
            streamInfo.cameraID = info.cameraID;
            streamInfo.streamIndex = info.streamIndex;
            streamInfo.pfStreamDataCB = streamDataCB;
            streamInfo.pfPCMDataCB = null;
            streamInfo.pfYV12DataCB = null;
            streamInfo.pfVideoDecodeCB = null;
            streamInfo.pfStreamStatusCB = streamStatusCB;
            streamInfo.hDisplayWindow = panelViewCell.Handle;
            streamInfo.capVersion = info.capVersion;
            streamInfo.useSSL = info.useSSL;
            streamInfo.videoDisplayMode = PlatformSDK.EVideoDisplayMode.fillViewCell;

            int ret = PlatformSDK.CreateLiveChannel(ref hChannel, ref streamInfo);
            EventLog.Write("PlatformSDK.CreateLiveChannel Ret = {0:X}", ret);

            if (ret != 0) return;
            
            SetDisplayInfo();

            buttonForwardOperation.Hide();
            buttonBackwardOperation.Hide();
            buttonSpeedUp.Hide();
            buttonSpeedDown.Hide();
            labelSpeed.Hide();

            channelType = EChannelType.live;
            streamState = EStreamState.play;               
        }

        void CreateDecodeLiveChannel(TCreateChannelInfo info)
        {
            TestLoopShowSnapshot();

            TStreamInfo streamInfo = new TStreamInfo();

            streamInfo.streamSource = PlatformSDK.EStreamSource.VAST;
            streamInfo.streamMode = PlatformSDK.EStreamMode.normalStream;
            streamInfo.serverIP = info.serverIP;
            streamInfo.serverUserName = info.serverUserName;
            streamInfo.serverPassword = info.serverPassword;
            streamInfo.serverPort = Convert.ToInt32(info.serverPort);
            streamInfo.cameraID = info.cameraID;
            streamInfo.streamIndex = info.streamIndex;
            streamInfo.pfStreamDataCB = streamDataCB;
            streamInfo.pfPCMDataCB = null;
            streamInfo.pfYV12DataCB = null;
            streamInfo.pfVideoDecodeCB = videoDecodeCB;
            streamInfo.pfStreamStatusCB = null;
            streamInfo.hDisplayWindow = IntPtr.Zero;
            streamInfo.capVersion = info.capVersion;
            streamInfo.useSSL = info.useSSL;
            streamInfo.videoDisplayMode = PlatformSDK.EVideoDisplayMode.fillViewCell;

            int ret = PlatformSDK.CreateLiveChannel(ref hChannel, ref streamInfo);
            EventLog.Write("PlatformSDK.CreateLiveChannel Ret = {0:X}", ret);

            if (ret != 0) return;

            SetDisplayInfo();

            buttonForwardOperation.Hide();
            buttonBackwardOperation.Hide();
            buttonSpeedUp.Hide();
            buttonSpeedDown.Hide();
            labelSpeed.Hide();

            channelType = EChannelType.live;
            streamState = EStreamState.play;
        }

        public void CreateLiveChannel(ref TCreateChannelInfo info)
        {
            DeleteLiveChannel();
            //CreateDecodeLiveChannel(info);
            CreateLiveChannel(info);
        }

        public void DeleteLiveChannel()
        {
            if (hChannel == IntPtr.Zero || channelType != EChannelType.live) return;
            int ret = PlatformSDK.DeleteLiveChannel(ref hChannel);
            EventLog.Write("PlatformSDK.DeleteLiveChannel Ret = {0:X}", ret);
        }

        void CreateNetFileChannel(TCreateChannelInfo info)
        {
            TStreamInfo streamInfo = new TStreamInfo();

            streamInfo.streamSource = PlatformSDK.EStreamSource.VAST;
            streamInfo.streamMode = PlatformSDK.EStreamMode.normalStream;
            streamInfo.timeType = PlatformSDK.ETimeType.recordingTime;
            streamInfo.serverIP = info.serverIP;
            streamInfo.serverUserName = info.serverUserName;
            streamInfo.serverPassword = info.serverPassword;
            streamInfo.serverPort = info.serverPort;
            streamInfo.cameraID = info.cameraID;
            streamInfo.pfStreamStatusCB = streamStatusCB;
            streamInfo.pfStreamDataCB = streamDataCB;
            streamInfo.pfPCMDataCB = null;
            streamInfo.pfYV12DataCB = null;
            streamInfo.pfVideoDecodeCB = null;
            streamInfo.hDisplayWindow = panelViewCell.Handle;
            streamInfo.startTime = info.startTime;
            streamInfo.endTime = info.endTime;
            streamInfo.streamIndex = info.streamIndex;
            streamInfo.streamInitStatus = PlatformSDK.EStreamInitStatus.play;
            streamInfo.timeKind = PlatformSDK.ETimeKind.UTC;
            streamInfo.capVersion = info.capVersion;
            streamInfo.useSSL = info.useSSL;
            streamInfo.videoDisplayMode = PlatformSDK.EVideoDisplayMode.fillViewCell;
            streamInfo.playbackMode = PlatformSDK.EPlaybackMode.normal;
            streamInfo.iFrameOnlyInterval = iFrameOnlyInterval;
            EventLog.Write("Set normal mode iFrameOnlyInterval= {0}", streamInfo.iFrameOnlyInterval);

            int ret = PlatformSDK.CreateNetFileChannel(ref hChannel, ref streamInfo);
            EventLog.Write("PlatformSDK.CreateNetFileChannel Ret = {0:X}", ret);

            if (ret != 0) return;

            SetDisplayInfo();

            buttonForwardOperation.Text = pauseText;
            channelType = EChannelType.playback;
            streamState = EStreamState.play;
        }
        void CreateNetFileChannel_RawData(TCreateChannelInfo info)
        {
            TStreamInfo streamInfo = new TStreamInfo();

            streamInfo.streamSource = PlatformSDK.EStreamSource.VAST;
            streamInfo.streamMode = PlatformSDK.EStreamMode.normalStream;
            streamInfo.timeType = PlatformSDK.ETimeType.recordingTime;
            streamInfo.serverIP = info.serverIP;
            streamInfo.serverUserName = info.serverUserName;
            streamInfo.serverPassword = info.serverPassword;
            streamInfo.serverPort = info.serverPort;
            streamInfo.cameraID = info.cameraID;
            streamInfo.pfStreamStatusCB = null;
            streamInfo.pfStreamDataCB = null;
            streamInfo.pfYV12DataCB = null;
            streamInfo.hDisplayWindow = IntPtr.Zero;
            streamInfo.pfRawDataCB = rawDataCallBack;
            streamInfo.rawDataFormat = PlatformSDK.ERawDataFormat.RGB24;

            streamInfo.startTime = info.startTime;
            streamInfo.endTime = info.endTime;
            streamInfo.streamIndex = info.streamIndex;
            streamInfo.streamInitStatus = PlatformSDK.EStreamInitStatus.play;
            streamInfo.timeKind = PlatformSDK.ETimeKind.UTC;
            streamInfo.capVersion = info.capVersion;
            streamInfo.useSSL = info.useSSL;
            streamInfo.videoDisplayMode = PlatformSDK.EVideoDisplayMode.fillViewCell;
            streamInfo.playbackMode = PlatformSDK.EPlaybackMode.normal;
            streamInfo.iFrameOnlyInterval = iFrameOnlyInterval;
            EventLog.Write("Set normal mode iFrameOnlyInterval= {0}, {1}", streamInfo.iFrameOnlyInterval, info.serverPort);

            int ret = PlatformSDK.CreateNetFileChannel(ref hChannel, ref streamInfo);
            EventLog.Write("PlatformSDK.CreateNetFileChannel Ret = {0:X}", ret);

            if (ret != 0) return;

            SetDisplayInfo();

            buttonForwardOperation.Text = pauseText;
            channelType = EChannelType.playback;
            streamState = EStreamState.play;
        }

        void CreateDecodeNetFileChannel(TCreateChannelInfo info)
        {
            TestLoopShowSnapshot();

            TStreamInfo streamInfo = new TStreamInfo();

            streamInfo.streamSource = PlatformSDK.EStreamSource.VAST;
            streamInfo.streamMode = PlatformSDK.EStreamMode.normalStream;
            streamInfo.timeType = PlatformSDK.ETimeType.recordingTime;
            streamInfo.serverIP = info.serverIP;
            streamInfo.serverUserName = info.serverUserName;
            streamInfo.serverPassword = info.serverPassword;
            streamInfo.serverPort = info.serverPort;
            streamInfo.cameraID = info.cameraID;
            streamInfo.pfStreamStatusCB = streamStatusCB;
            streamInfo.pfStreamDataCB = streamDataCB;
            streamInfo.pfPCMDataCB = null;
            streamInfo.pfYV12DataCB = null;
            streamInfo.pfVideoDecodeCB = videoDecodeCB;
            streamInfo.hDisplayWindow = IntPtr.Zero;
            streamInfo.startTime = info.startTime;
            streamInfo.endTime = info.endTime;
            streamInfo.streamIndex = info.streamIndex;
            streamInfo.streamInitStatus = PlatformSDK.EStreamInitStatus.play;
            streamInfo.timeKind = PlatformSDK.ETimeKind.UTC;
            streamInfo.capVersion = info.capVersion;
            streamInfo.useSSL = info.useSSL;
            streamInfo.videoDisplayMode = PlatformSDK.EVideoDisplayMode.fillViewCell;

            int ret = PlatformSDK.CreateNetFileChannel(ref hChannel, ref streamInfo);
            EventLog.Write("PlatformSDK.CreateNetFileChannel Ret = {0:X}", ret);

            if (ret != 0) return;

            SetDisplayInfo();

            buttonForwardOperation.Text = pauseText;
            channelType = EChannelType.playback;
            streamState = EStreamState.play;

            buttonForwardOperation.Hide();
            buttonBackwardOperation.Hide();
            buttonSpeedUp.Hide();
            buttonSpeedDown.Hide();
            labelSpeed.Hide();
        }

        public void CreateNetFileChannel(ref TCreateChannelInfo info)
        {
            DeleteNetFileChannel();
            //CreateDecodeNetFileChannel(info);
            CreateNetFileChannel(info);
        }

        public void CreateNetFileChannel_RawData(ref TCreateChannelInfo info)
        {
            DeleteNetFileChannel();
            CreateNetFileChannel_RawData(info);
        }

        public IntPtr CreateSyncChannel(ref TCreateChannelInfo info)
        {
            TStreamInfo streamInfo = new TStreamInfo();

            streamInfo.streamSource = PlatformSDK.EStreamSource.VAST;
            streamInfo.streamMode = PlatformSDK.EStreamMode.normalStream;
            streamInfo.timeType = PlatformSDK.ETimeType.recordingTime;
            streamInfo.serverIP = info.serverIP;
            streamInfo.serverUserName = info.serverUserName;
            streamInfo.serverPassword = info.serverPassword;
            streamInfo.serverPort = info.serverPort;
            streamInfo.cameraID = info.cameraID;
            streamInfo.pfStreamStatusCB = streamStatusCB;
            streamInfo.pfStreamDataCB = streamDataCB;
            streamInfo.pfPCMDataCB = null;
            streamInfo.pfYV12DataCB = null;
            streamInfo.pfVideoDecodeCB = null;
            streamInfo.hDisplayWindow = panelViewCell.Handle;
            streamInfo.startTime = info.startTime;
            streamInfo.endTime = info.endTime;
            streamInfo.streamIndex = info.streamIndex;
            streamInfo.streamInitStatus = PlatformSDK.EStreamInitStatus.play;
            streamInfo.timeKind = PlatformSDK.ETimeKind.UTC;
            streamInfo.capVersion = info.capVersion;
            streamInfo.useSSL = info.useSSL;
            streamInfo.videoDisplayMode = PlatformSDK.EVideoDisplayMode.fillViewCell;
            streamInfo.playbackMode = PlatformSDK.EPlaybackMode.synchronous;
            streamInfo.iFrameOnlyInterval = iFrameOnlyInterval;
            EventLog.Write("Set sync mode iFrameOnlyInterval= {0}", streamInfo.iFrameOnlyInterval);

            int ret = PlatformSDK.CreateNetFileChannel(ref hChannel, ref streamInfo);
            EventLog.Write("PlatformSDK.CreateNetFileChannel Ret = {0:X}", ret);

            SetDisplayInfo();

            buttonForwardOperation.Text = pauseText;
            channelType = EChannelType.playback;
            streamState = EStreamState.play;
            startTime = info.startTime;
            endTime = info.endTime;

            isSyncPlayback = true;

            return hChannel;
        }

        public void DeleteNetFileChannel()
        {
            if (hChannel == IntPtr.Zero || channelType != EChannelType.playback) return;
            int ret = PlatformSDK.DeleteNetFileChannel(ref hChannel);
            EventLog.Write("PlatformSDK.DeleteNetFileChannel Ret = {0:X}", ret);

            timerLoopShowSnapshot.Enabled = false;
        }

        void CreateLocalFileChannel(ref TCreateChannelInfo chInfo)
        {
            TLocalFileInfo info = new TLocalFileInfo();
            info.filePath = chInfo.filePath;
            info.hDisplayWindow = panelViewCell.Handle;
            int ret = PlatformSDK.CreateLocalFileChannel(ref hChannel, ref info);
            EventLog.Write("PlatformSDK.CreateLocalFileChannel Ret = {0:X}", ret);

            if (ret != 0) return;

            SetDisplayInfo();

            buttonForwardOperation.Text = pauseText;
            channelType = EChannelType.localFile;
            streamState = EStreamState.play;
        }

        public void CreateLocalFileChannel(string filePath)
        {
            reCreateChannelInfo.filePath = filePath;
            CreateLocalFileChannel(ref reCreateChannelInfo);
        }

        public void DeleteLocalFileChannel()
        {
            if (hChannel == IntPtr.Zero || channelType != EChannelType.localFile) return;
            int ret = PlatformSDK.DeleteLocalFileChannel(ref hChannel);
            EventLog.Write("PlatformSDK.DeleteLocalFileChannel Ret = {0:X}", ret);
        }

        void DeleteChannel()
        {
            if (channelType == EChannelType.live)
            {
                DeleteLiveChannel();
            }
            else if (channelType == EChannelType.playback)
            {
                DeleteNetFileChannel();
            }
            else if (channelType == EChannelType.localFile)
            {
                DeleteLocalFileChannel();
            }

            streamState = EStreamState.stop;
        }

        private void ViewCell_Closed(object sender, EventArgs e)
        {
            DeleteChannel();  
        }

        void ReCreateChannel()
        {
            if (channelType == EChannelType.live)
            {
                CreateLiveChannel(ref reCreateChannelInfo);
            }
            else if (channelType == EChannelType.playback)
            {
                CreateNetFileChannel(ref reCreateChannelInfo);
            }
            else if (channelType == EChannelType.localFile)
            {
                CreateLocalFileChannel(ref reCreateChannelInfo);
            }
        }

        void StartChannel()
        {
            int ret = PlatformSDK.StartChannel(hChannel);
            Console.WriteLine("PlatformSDK.StartChannel Ret = {0:X}", ret);
            buttonForwardOperation.Text = pauseText;
            streamState = EStreamState.play;
        }

        void PauseChannel()
        {
            int ret = PlatformSDK.PauseChannel(hChannel);
            Console.WriteLine("PlatformSDK.PauseChannel Ret = {0:X}", ret);
            buttonForwardOperation.Text = forwardPlayText;
            streamState = EStreamState.pause;
        }

        void HandleLiveOption()
        {
            if (streamState == EStreamState.play)
            {
                DeleteChannel();
                buttonStop.Text = forwardPlayText;
            }
            else if (streamState == EStreamState.stop)
            {
                ReCreateChannel();
                buttonStop.Text = stopText;
            }
        }

        void HandleOperation()
        {
            if (streamState == EStreamState.pause)
            {
                StartChannel();
            }
            else if (streamState == EStreamState.play)
            {
                PauseChannel();
            }
            else if (streamState == EStreamState.stop)
            {
                ReCreateChannel();
            }
        }

        private void buttonForwardOperation_Click(object sender, EventArgs e)
        {
            playDirection = EPlayDirection.forward;
            HandleOperation();
        }

        void ChangeSpeed()
        {
            int ret = PlatformSDK.ChangeSpeed(hChannel, speed);
            EventLog.Write("PlatformSDK.ChangeSpeed Ret = {0:X} speed {1}", ret, speed);
            
            if (ret != 0) return;
            labelSpeed.Text = speed.ToString("0.000");
        }

        private void buttonSpeedUp_Click(object sender, EventArgs e)
        {            
            speed *= 2;
            if (speed >= MAX_SPEED)
            {
                speed = MAX_SPEED;
                return;
            }
            ChangeSpeed();
        }

        private void buttonSpeedDown_Click(object sender, EventArgs e)
        {            
            speed /= 2;
            if (speed <= MIN_SPEED)
            {
                speed = MIN_SPEED;
                return;
            }
            ChangeSpeed();
        }

        private void buttonStop_Click(object sender, EventArgs e)
        {
            if (channelType == EChannelType.live)
            {
                HandleLiveOption();
            }
            else
            {
                DeleteChannel();
            }            
        }

        private void buttonBackWardOperation_Click(object sender, EventArgs e)
        {
            playDirection = EPlayDirection.backward;
            int ret = PlatformSDK.UpdatePlaybackDirection(hChannel, EPlayDirection.backward);
            EventLog.Write("PlatformSDK.UpdatePlaybackDirection to backward Ret = {0:X}", ret);
        }

        private void menuViewOption_ItemClicked(object sender, ToolStripItemClickedEventArgs e)
        {
            ToolStripItem item = e.ClickedItem;
        }

        private void buttonFE1R_Click(object sender, EventArgs e)
        {
            PlatformSDK.SetFisheyeDisplayMode(hChannel, PlatformSDK.EFisheyeDisplayMode.mode_1R);
        }

        private void buttonFE1O_Click(object sender, EventArgs e)
        {
            PlatformSDK.SetFisheyeDisplayMode(hChannel, PlatformSDK.EFisheyeDisplayMode.mode_1O);
        }

        private void ViewCell_GotFocus(object sender, EventArgs e)
        {
            PlatformSDK.AudioOn(hChannel, true);
        }

        private void ViewCell_LostFocus(object sender, EventArgs e)
        {
            PlatformSDK.AudioOn(hChannel, false);
        }

        private void snapshotButton_Click(object sender, EventArgs e)
        {
            IntPtr path = Marshal.StringToHGlobalAnsi("D:\\");
            IntPtr fileName = Marshal.StringToHGlobalAnsi("Test");

            PlatformSDK.TChannelSnapshotInfo channelSnapshotInfo = new PlatformSDK.TChannelSnapshotInfo();
            channelSnapshotInfo.pszOutputDir = path;
            channelSnapshotInfo.pszFileName = fileName;

            var ret = PlatformSDK.SaveChannelSnapshot(hChannel, ref channelSnapshotInfo);
            EventLog.Write("PlatformSDK.SaveChannelSnapshot Ret = {0:X}", ret);
        }

        private void reconnectButton_Click(object sender, EventArgs e)
        {
            var ret = PlatformSDK.StartChannel(hChannel);
            EventLog.Write("PlatformSDK.StartChannel Ret = {0:X}", ret);
        }

        private void ToggleAudio_Click(object sender, EventArgs e)
        {
            turn = !turn;
            var ret = PlatformSDK.AudioOn(hChannel, turn);
            EventLog.Write("PlatformSDK.AudioOn Ret = {0:X}", ret);
        }

        private void getSnapshot_Click(object sender, EventArgs e)
        {
            PlatformSDK.TSnapshotRawInfo snapshotRawInfo = new PlatformSDK.TSnapshotRawInfo();

            var ret = PlatformSDK.GetChannelSnapshot(hChannel, ref snapshotRawInfo);
            EventLog.Write("PlatformSDK.GetChannelSnapshot Ret = {0:X}", ret);

            if (snapshotRawInfo.pBuffer == IntPtr.Zero) return;
            
            string savePath = "D:\\FromRawData.jpg";
            try
            {
                byte[] image = new byte[snapshotRawInfo.bufSize];
                Marshal.Copy(snapshotRawInfo.pBuffer, image, 0, snapshotRawInfo.bufSize);
                FileStream file = new FileStream(savePath, FileMode.OpenOrCreate, FileAccess.Write);
                file.Write(image, 0, snapshotRawInfo.bufSize);
                file.Close();
            }
            catch (IOException err)
            {
                EventLog.Write("Fail to save image \"" + savePath + "\":\n" + err.ToString());
            }
        }
    }

    public class TCreateChannelInfo
    {
        public string serverIP;
        public string serverUserName;
        public string serverPassword;
        public int serverPort;
        public string cameraID;
        public short streamIndex;
        public uint startTime;
        public uint endTime;
        public string filePath;
        public ECapabilityVersion capVersion;
        public bool useSSL;        
    };
}
